(** Common BSP's functions *)

open Tools
open Zarith
open Bsmllib
open Bsmlutils

(** For the functions, see the Bsmllib's reference manuel *)

let replicate a =
  mkpar (fun pid -> a)

let parfun f v =
  apply (replicate f) v

let parfun2 f v1 v2 =
  apply (parfun f v1) v2

let parfun3 f v1 v2 v3 =
  apply (parfun2 f v1 v2) v3

let parfun4 f v1 v2 v3 v4 =
  apply (parfun3 f v1 v2 v3) v4

let apply2 f v1 v2 =
  apply (apply f v1) v2

let apply3 f v1 v2 v3 =
  apply (apply2 f v1 v2) v3

let apply4 f v1 v2 v3 v4 =
  apply (apply3 f v1 v2 v3) v4

let this () =
  mkpar (fun pid -> pid)
  
let last () =
  zminus (bsp_p ()) (POS XH)

let none x =
  None

let within_bound i =
  match z_le_dec ZERO i with
    | true -> z_lt_dec i (bsp_p ())
    | false -> false

let noSome = function
  | Some t -> t
  | None -> bsp_FAIL ()

let hd = function
  | Nil -> bsp_FAIL ()
  | Cons (t, l0) -> t

let tl = function
  | Nil -> bsp_FAIL ()
  | Cons (t, l0) -> l0

let procs h =
  from_to (last ())

let pair_of_par arg =
  Pair ((apply (replicate (fun x -> fst x)) arg),
    (apply (replicate (fun x -> snd x)) arg))

let par_of_pair arg =
  apply2 (replicate (fun xi yi -> Pair (xi, yi))) (fst arg) (snd arg)

let mask c x y =
  apply
    (apply
      (mkpar (fun j x0 y0 ->
        match sumbool_of_bool (c j) with
          | true -> x0
          | false -> y0)) x) y

let noSome_vect v =
  parfun (fun x -> noSome x) v

let tl_hd_vect v =
  parfun (fun l -> Pair ((hd l), (tl l))) v

let no_some_in_list_vect v =
  parfun (fun x -> map (fun x0 -> noSome x0) x) v

let get_one_tmp datas pids =
  apply
    (put
      (parfun2 (fun f d dst ->
        match f dst with
          | Some y -> Some d
          | None -> None)
        (put
          (parfun (fun i dst ->
            match z_eq_dec dst i with
              | true -> Some ()
              | false -> None) pids)) datas)) pids

let get_one datas pids =
  noSome_vect (get_one_tmp datas pids)

let get_list_tmp datas lpids =
  parfun2 (fun x x0 -> map x x0)
    (put
      (parfun2 (fun f d dst ->
        match f dst with
          | Some u -> Some d
          | None -> None)
        (put
          (parfun (fun l dst ->
            match let rec f = function
                    | Nil -> false
                    | Cons (a, l1) ->
                        (match z_eq_dec a dst with
                           | true -> true
                           | false -> f l1)
                  in f l with
              | true -> Some ()
              | false -> None) lpids)) datas)) lpids

let get_list datas lpids =
  no_some_in_list_vect (get_list_tmp datas lpids)

let bsml_print p at_pid vec =
  match within_bound at_pid with
    | true ->
        apply
          (mkpar (fun pid ->
            match z_eq_dec pid at_pid with
              | true -> p
              | false -> (fun x -> ()))) vec
    | false -> assert false (* absurd case *)

let parprint p vec =
  parfun2 (fun pid data -> p data) (this ()) vec

(** Functions for the parallel superposition *)

let inbounds first last0 n =
  match z_le_dec first n with
    | true -> z_le_dec n last0
    | false -> false

let mix m v =
  apply
    (apply
      (mkpar (fun pid v1 v2 ->
        match z_le_dec pid m with
          | true -> v1
          | false -> v2)) (fst v)) (snd v)
