(** Common BSP's functions of communication *)

open Zarith
open Bsmllib
open Bsmlbase
open Bsmlutils
open Tools

let shift_tmp dec datas =
  apply (put (parfun2 (fun data pid dst -> Some data) datas (this ())))
    (mkpar (fun pid ->
      zmod (zminus (zplus (bsp_p ()) pid) (zmod dec (bsp_p ())))
        (bsp_p ())))

(** General shift parallel vector *)
let shift dec datas =
  noSome_vect (shift_tmp dec datas)

(** Shift left *)
let shift_left data =
  shift (POS XH) data

(** Shift right *)
let shift_right data =
  shift (NEG XH) data


let totex_tmp vec =
  put (parfun (fun data dst -> Some data) vec)


(** Totex *)
let totex vec =
  parfun (fun f src -> noSome (f src)) (totex_tmp vec)

(** total exchange *)
let total_exchange vec =
  parfun2 (fun x x0 -> map x x0) (totex vec) (replicate (procs ()))

let bcast_direct_tmp root vv =
  apply
    (put
      (apply
        (mkpar (fun pid v dst ->
          match z_eq_dec pid root with
            | true -> Some v
            | false -> None)) vv)) (replicate root)


(** Broadcast direct *)
let bcast_direct root vv =
  match within_bound root with
    | true -> noSome_vect (bcast_direct_tmp root vv)
    | false -> assert false (* absurd case *)


let scatter_tmp partition root v =
  apply
    (put
      (apply
        (mkpar (fun pid ->
          match z_eq_dec pid root with
            | true -> partition
            | false -> (fun a x -> none x))) v)) (replicate root)

(** Scatter a vector with partition from root*)
let scatter partition root v =
  match within_bound root with
    | true -> noSome_vect (scatter_tmp partition root v)
    | false -> assert false (* absurd case *)

(** Gather a parallel vector *)
let gather root vv =
  put
    (apply
      (mkpar (fun pid v dest ->
        match z_eq_dec dest root with
          | true -> Some v
          | false -> None)) vv)


let procs_lists_gather root =
  mkpar (fun pid ->
    match z_eq_dec root pid with
      | true -> from_to2 (zminus (bsp_p ()) (POS XH))
      | false -> Nil)

(** Gather a list of values *)
let gather_list vec root =
  parfun2 (fun x x0 -> map x x0)
    (parfun (fun x x0 -> compose (fun x1 -> noSome x1) x x0)
      (gather root vec)) (procs_lists_gather root)


let procs_lists h =
  mkpar (fun pid -> from_to2 pid)

let scan_direct_tmp1 vec =
  parfun2 (fun x x0 -> map x x0)
    (parfun (fun x x0 -> compose (fun x1 -> noSome x1) x x0)
      (put
        (apply
          (mkpar (fun pid v dst ->
            match z_lt_dec dst pid with
              | true -> None
              | false -> Some v)) vec))) (procs_lists ())

(** Scan Direct *)
let scan_direct op0 vec =
  parfun (fun hd_tl -> fold_left op0 (snd hd_tl) (fst hd_tl))
    (tl_hd_vect (scan_direct_tmp1 vec))

(** Fold_left *)
let fold_left_vec op0 vec =
  parfun (fun hd_tl -> fold_left op0 (snd hd_tl) (fst hd_tl))
    (tl_hd_vect (total_exchange vec))
