(** A Sequential Simulo for the Certified Library *)

open Zarith
open Tools

(** Number of processes *)
let bsp_p () =
  POS (XO (XO (XO (XO XH))))

(** Abstract type of the parallel vector *)
type 'a par = Par of 'a array


(** Simulo of asubset of the Array module for the type "Z"  *)

(** Array.init *)
let array_init n f = 
  let array_f = fun x -> (f (int_to_z x)) in
    Array.init (z_to_int n) array_f

(** tab.(n) *)
let array_point tab n = tab.(z_to_int n)



(**  The parallel operators  *)

(** Create a parallel vector *)
let mkpar f =Par(array_init (bsp_p()) f) 

(** Asynchronous application *)
let apply (Par vf) (Par vv) =
  Par (array_init (bsp_p()) (fun i-> (array_point vf (i)) (array_point vv (i))))

(** Function for the "if at" construction *)
let at (Par (vb:bool array)) n =  array_point vb n

(** Exchange of values *)
let put (Par (vf:(z->'a option) array)) = 
  Par(array_init (bsp_p()) (fun i -> fun j -> 
           if (z_le_dec ZERO j) & (z_lt_dec j (bsp_p())) then ((array_point vf j) i) else None))

(** Super-thread in BSP computation *)	   
let super (vv1:(unit -> ('a par)))  (vv2:(unit -> ('b par))) = (Pair (vv1 (),vv2 ()))

(** Error of one of the processor *)
let bsp_FAIL () = failwith "BSP_FAIL"
