(** Common sequential and usefull functions *)

open Zarith
open Tools

(** Type of the list *)
type 'a list =
  | Nil
  | Cons of 'a * 'a list


(** Length of a list *)
let rec zlength = function
  | Nil -> ZERO
  | Cons (t, m) -> zplus (zlength m) (POS XH)
  

(** Nth element of a list *)
let rec znth l pos =
  match l with
    | Nil -> assert false (* absurd case *)
    | Cons (a, l0) ->
        (match z_eq_dec pos (POS XH) with
           | true -> a
           | false -> znth l0 (zminus pos (POS XH)))

(** Concaten lists *)
let rec app l m =
  match l with
    | Nil -> m
    | Cons (a, l1) -> Cons (a, (app l1 m))

(** Map a list with a function *)
let rec map f = function
  | Nil -> Nil
  | Cons (a, t) -> Cons ((f a), (map f t))

(** Fold direct *)
let rec fold_left f x x0 =
  match x with
    | Nil -> x0
    | Cons (b, t) -> fold_left f t (f x0 b)   	   


(** For these functions see the BSmllib's reference manuel *) 

let mklist x =
  Cons (x, Nil)

let id x =
  x

let compose f g x =
  f (g x)

let is_Some = function
  | Some t -> true
  | None -> false 


let from_to_F n fromto_recf =
  match z_eq_dec ZERO n with
    | true -> Cons (ZERO, Nil)
    | false ->
        app (fromto_recf (zminus n (POS XH)) () ()) (Cons (n, Nil))

(** From_to *)
let rec from_to x =
  from_to_F x (fun y _ _ -> from_to y)

(** From_to with absolue *)
let from_to2 n =
  from_to (zabs n)

(** filtermap *)
let rec filtermap l p f =
  match l with
    | Nil -> Nil
    | Cons (a, l0) ->
        let x = filtermap l0 p f in
        (match sumbool_of_bool (p a) with
           | true -> Cons ((f a), x)
           | false -> x)

(** to_list *)
let rec to_list l f =
  match l with
    | Nil -> Nil
    | Cons (a, l0) ->
        let x = to_list l0 f in
        (match option_dec (f a) with
           | Some s -> Cons (s, x)
           | None -> x)
