(** A Subset of the Coq's Certified Arithmetic Library *)

(** For other information, see the Coq's library *)

(* Type of the positive *)
type positive =
  | XI of positive
  | XO of positive
  | XH

(** type of z *)  
type z =
  | ZERO
  | POS of positive
  | NEG of positive

(* type of natural number *)
type entier =
  | Nul
  | Pos of positive

(* type of a relation *)
type relation =
  | EGAL
  | INFERIEUR
  | SUPERIEUR

(* Auxiliary functions *)  
let zero_suivi_de = function
  | Nul -> Nul
  | Pos p -> Pos (XO p)

let un_suivi_de = function
  | Nul -> Pos XH
  | Pos p -> Pos (XI p)

let rec double_moins_un = function
  | XI x' -> XI (XO x')
  | XO x' -> XI (double_moins_un x')
  | XH -> XH

let double_moins_deux = function
  | XI x' -> Pos (XO (XO x'))
  | XO x' -> Pos (XO (double_moins_un x'))
  | XH -> Nul

let rec sub_pos x y =
  match x with
    | XI x' ->
        (match y with
           | XI y' -> zero_suivi_de (sub_pos x' y')
           | XO y' -> un_suivi_de (sub_pos x' y')
           | XH -> Pos (XO x'))
    | XO x' ->
        (match y with
           | XI y' -> un_suivi_de (sub_neg x' y')
           | XO y' -> zero_suivi_de (sub_pos x' y')
           | XH -> Pos (double_moins_un x'))
    | XH ->
        (match y with
           | XI y' -> Pos (double_moins_un y')
           | XO y' -> double_moins_deux y'
           | XH -> Nul)
and sub_neg x y =
  match x with
    | XI x' ->
        (match y with
           | XI y' -> un_suivi_de (sub_neg x' y')
           | XO y' -> zero_suivi_de (sub_pos x' y')
           | XH -> Pos (double_moins_un x'))
    | XO x' ->
        (match y with
           | XI y' -> zero_suivi_de (sub_neg x' y')
           | XO y' -> un_suivi_de (sub_neg x' y')
           | XH -> double_moins_deux x')
    | XH ->
        (match y with
           | XI y' -> Pos (XO y')
           | XO y' -> Pos (double_moins_un y')
           | XH -> Nul)


let rec compare x y r =
  match x with
    | XI x' ->
        (match y with
           | XI y' -> compare x' y' r
           | XO y' -> compare x' y' SUPERIEUR
           | XH -> SUPERIEUR)
    | XO x' ->
        (match y with
           | XI y' -> compare x' y' INFERIEUR
           | XO y' -> compare x' y' r
           | XH -> SUPERIEUR)
    | XH ->
        (match y with
           | XI y' -> INFERIEUR
           | XO y' -> INFERIEUR
           | XH -> r)
 
 
let dcompare_inf = function
  | EGAL -> Some true
  | INFERIEUR -> Some false
  | SUPERIEUR -> None


let rec add_un = function
  | XI x' -> XO (add_un x')
  | XO x' -> XI x'
  | XH -> XO XH

let rec add x y =
  match x with
    | XI x' ->
        (match y with
           | XI y' -> XO (add_carry x' y')
           | XO y' -> XI (add x' y')
           | XH -> XO (add_un x'))
    | XO x' ->
        (match y with
           | XI y' -> XI (add x' y')
           | XO y' -> XO (add x' y')
           | XH -> XI x')
    | XH ->
        (match y with
           | XI y' -> XO (add_un y')
           | XO y' -> XI y'
           | XH -> XO XH)
and add_carry x y =
  match x with
    | XI x' ->
        (match y with
           | XI y' -> XI (add_carry x' y')
           | XO y' -> XO (add_carry x' y')
           | XH -> XI (add_un x'))
    | XO x' ->
        (match y with
           | XI y' -> XO (add_carry x' y')
           | XO y' -> XI (add x' y')
           | XH -> XO (add_un x'))
    | XH ->
        (match y with
           | XI y' -> XI (add_un y')
           | XO y' -> XO (add_un y')
           | XH -> XI XH)

let rec times1 x f y =
  match x with
    | XI x' -> add (f y) (times1 x' (fun z -> XO (f z)) y)
    | XO x' -> times1 x' (fun z -> XO (f z)) y
    | XH -> f y



(* Opposite relation *)
let op = function
  | EGAL -> EGAL
  | INFERIEUR -> SUPERIEUR
  | SUPERIEUR -> INFERIEUR


(* Comparaison in Z *) 
let zcompare x y =
  match x with
    | ZERO ->
        (match y with
           | ZERO -> EGAL
           | POS y' -> INFERIEUR
           | NEG y' -> SUPERIEUR)
    | POS x' ->
        (match y with
           | ZERO -> SUPERIEUR
           | POS y' -> compare x' y' EGAL
           | NEG y' -> SUPERIEUR)
    | NEG x' ->
        (match y with
           | ZERO -> INFERIEUR
           | POS y' -> INFERIEUR
           | NEG y' -> op (compare x' y' EGAL))
  
let zcompare_rec x y h1 h2 h3 =
  match dcompare_inf (zcompare x y) with
    | Some x0 -> (match x0 with
                      | true -> h1 ()
                      | false -> h2 ())
    | None -> h3 ()


(** <= in Z *)
let z_le_dec x y =
  zcompare_rec x y (fun _ -> true) (fun _ -> true) (fun _ -> false)

(** < in Z *)
let z_lt_dec x y =
  zcompare_rec x y (fun _ -> false) (fun _ -> true) (fun _ -> false)

(** = in Z *)  
let z_eq_dec x y =
  zcompare_rec x y (fun _ -> true) (fun _ -> false) (fun _ -> false)

(** >= in Z *)
let z_ge_dec x y =
  zcompare_rec x y (fun _ -> true) (fun _ -> false) (fun _ -> true)

(** > in Z *)
let z_gt_dec x y =
  zcompare_rec x y (fun _ -> false) (fun _ -> false) (fun _ -> true)



(* ************************ Addition in Z **************************** *)
(** Addition in Z *)
let zplus x x0 =
  let true_sub = fun x1 y ->
    match sub_pos x1 y with
      | Nul -> XH
      | Pos z0 -> z0
  in
  (match x with
     | ZERO -> x0
     | POS x' ->
         (match x0 with
            | ZERO -> x
            | POS y' -> POS (add x' y')
            | NEG y' ->
                (match compare x' y' EGAL with
                   | EGAL -> ZERO
                   | INFERIEUR -> NEG (true_sub y' x')
                   | SUPERIEUR -> POS (true_sub x' y')))
     | NEG x' ->
         (match x0 with
            | ZERO -> x
            | POS y' ->
                (match compare x' y' EGAL with
                   | EGAL -> ZERO
                   | INFERIEUR -> POS (true_sub y' x')
                   | SUPERIEUR -> NEG (true_sub x' y'))
            | NEG y' -> NEG (add x' y')))

(* ************************** Opposite in Z ****************************** *)
(** Opposite in Z *)
let zopp = function
  | ZERO -> ZERO
  | POS x0 -> NEG x0
  | NEG x0 -> POS x0
  
(* ******************************** Substraction in Z ********************* *)
(** Substrction in Z *)
let zminus m n = zplus m (zopp n)


 (* **************************** Multiplication in Z ********************* *)   
 (** Multiplication in Z *)
let zmult x x0 =
  let times = fun x1 -> times1 x1 (fun y -> y) in
  (match x with
     | ZERO -> ZERO
     | POS x' ->
         (match x0 with
            | ZERO -> ZERO
            | POS y' -> POS (times x' y')
            | NEG y' -> NEG (times x' y'))
     | NEG x' ->
         (match x0 with
            | ZERO -> ZERO
            | POS y' -> NEG (times x' y')
            | NEG y' -> POS (times x' y')))


(* Auxiliary function to Zdiv *)
let zgt_bool x y =
  match zcompare x y with
    | EGAL -> false
    | INFERIEUR -> false
    | SUPERIEUR -> true

let zge_bool x y =
  match zcompare x y with
    | EGAL -> true
    | INFERIEUR -> false
    | SUPERIEUR -> true

let rec zdiv_eucl_POS a b =
  match a with
    | XI a' ->
        let Tools.Pair (q, r) = zdiv_eucl_POS a' b in
        (match zgt_bool b (zplus (zplus r r) (POS XH)) with
           | true -> Tools.Pair ((zplus q q), (zplus (zplus r r) (POS XH)))
           | false -> Tools.Pair ((zplus (zplus q q) (POS XH)),
               (zminus (zplus (zplus r r) (POS XH)) b)))
    | XO a' ->
        let Tools.Pair (q, r) = zdiv_eucl_POS a' b in
        (match zgt_bool b (zplus r r) with
           | true -> Tools.Pair ((zplus q q), (zplus r r))
           | false -> Tools.Pair ((zplus (zplus q q) (POS XH)),
               (zminus (zplus r r) b)))
    | XH ->
        (match zge_bool b (POS (XO XH)) with
           | true -> Tools.Pair (ZERO, (POS XH))
           | false -> Tools.Pair ((POS XH), ZERO))


let zdiv_eucl a b =
  match a with
    | ZERO -> Tools.Pair (ZERO, ZERO)
    | POS a' ->
        (match b with
           | ZERO -> Tools.Pair (ZERO, ZERO)
           | POS p -> zdiv_eucl_POS a' b
           | NEG b' ->
               let Tools.Pair (q, r) = zdiv_eucl_POS a' (POS b') in
               (match r with
                  | ZERO -> Tools.Pair ((zopp q), ZERO)
                  | POS p -> Tools.Pair ((zopp (zplus q (POS XH))),
                      (zplus b r))
                  | NEG p -> Tools.Pair ((zopp (zplus q (POS XH))),
                      (zplus b r))))
    | NEG a' ->
        (match b with
           | ZERO -> Tools.Pair (ZERO, ZERO)
           | POS p ->
               let Tools.Pair (q, r) = zdiv_eucl_POS a' b in
               (match r with
                  | ZERO -> Tools.Pair ((zopp q), ZERO)
                  | POS p0 -> Tools.Pair ((zopp (zplus q (POS XH))),
                      (zminus b r))
                  | NEG p0 -> Tools.Pair ((zopp (zplus q (POS XH))),
                      (zminus b r)))
           | NEG b' ->
               let Tools.Pair (q, r) = zdiv_eucl_POS a' (POS b') in
               Tools.Pair (q, (zopp r)))

(* *************************** Division in Z *********************** *)
(** Division in Z *)
let zdiv a b =
  let Tools.Pair (q, x) = zdiv_eucl a b in q


(* ************************** Modulo in Z *************************** *)
(** Modulo in Z *)
let zmod a b =
  let Tools.Pair (x, r) = zdiv_eucl a b in r



(* Auxiliary function to zpower *)
let rec iter_pos n f x =
  match n with
    | XI n' -> f (iter_pos n' f (iter_pos n' f x))
    | XO n' -> iter_pos n' f (iter_pos n' f x)
    | XH -> f x
    
let zpower_pos z0 n =
  iter_pos n (fun x -> zmult z0 x) (POS XH)

(* ************************** Power in Z *************************** *)
(** Power in Z *)
let zpower x = function
  | ZERO -> POS XH
  | POS p -> zpower_pos x p
  | NEG p -> ZERO

(* ************************** Absolue in Z *************************** *)
(** Absolue in Z *)
let zabs = function
  | ZERO -> ZERO
  | POS p -> POS p
  | NEG p -> POS p


(* ***** Transform type: Coq' "Z" to Objective Caml's "int" ***** *)
let rec positive_to_int my_p pow2 = match my_p with
			  XI p -> pow2+(positive_to_int p (2*pow2))
			| XO p -> (positive_to_int p (2*pow2))
			| XH   -> pow2

(** Transform Coq' "Z" to Objective Caml's "int"  *)
let z_to_int my_z= match my_z with
		           ZERO  -> 0
			 | POS p -> (positive_to_int p 1)
			 | NEG p -> -(positive_to_int p 1)


(* ***** Transform type: Objective Caml's "int" to Coq' "Z" ***** *)

let rec int_to_positive = function
  | 0 -> XH
  | x -> add_un (int_to_positive (x-1))

(** Transform Objective Caml's   "int" to Coq' "Z" *) 
let int_to_z n = if (n=0) then ZERO else
                 if (n>0) then POS (int_to_positive (n-1)) else
		 NEG (int_to_positive ((-n)-1))
