(**  This module describe the axiomatization of parallel operator of the BSML language. Those operator
are given by parameters and their works in some axioms. *)

Require ZArith.
Require Bool.
Require Bool.Sumbool.
Require Omega.

Require Export ZArith.
Require Export Bool.
Require Export Bool.Sumbool.
Require Export Omega.

(** ** The BSML language *)

(** Parameters given by the BSMLlib library *)

Parameters bsp_p:unit -> Z;
    Vector : Set -> Set;
    at: (T:Set) (Vector T) -> (i:Z) `0<=i<(bsp_p tt)` -> T;
    mkpar  : (T:Set) (Z -> T) -> (Vector T);
    apply  : (T1,T2:Set) (Vector (T1 -> T2)) 
                            -> (Vector T1) -> (Vector T2);
    put    : (T:Set)(Vector (Z -> (option T))) 
                                   -> (Vector (Z -> (option T)));
    ifat   : (T:Set) (Vector bool) -> Z -> T -> T -> T;
    super  : (T1,T2:Set) (unit -> (Vector T1)) 
      -> (unit -> (Vector T2))  -> ((Vector T1)*(Vector T2)).


(** ** The formal axioms *)
(** Axiom for the number of processes  *)
Axiom good_bsp_p:`0 < (bsp_p tt)`.

(** function which given if "i" is a valid number of process *)
Definition within_bound: (i:Z) {`0<=i<(bsp_p tt)`}+{~`0<=i<(bsp_p tt)`}.
Intros; Case (Z_le_dec `0` i); Intros.
Intros; Case (Z_lt_dec i (bsp_p tt)); Intros; Auto.
Intuition.
Intuition.
Defined.


(** Axiom of the asynchronous operators *)
Axiom mkpar_def: (T:Set) (f:Z -> T)(i:Z) (H:`0 <= i < (bsp_p tt)`)
                    (at T (mkpar T f) i H)=(f i).
 
Axiom apply_def: (T1,T2:Set) (V1:(Vector (T1 -> T2))) 
                 (V2:(Vector T1 )) (i:Z) (H:`0 <= i < (bsp_p tt)`)
(at T2 (apply T1 T2 V1 V2) i H)=( (at (T1->T2) V1 i H) (at T1 V2 i H)).


(** Axiom of the communication-synchronisation operator *)
Axiom put_def: (T:Set)(Vf:(Vector (Z -> (option T)))) 
               (i:Z) (H:`0 <= i < (bsp_p tt)`)
((at (Z->(option T)) (put T Vf) i H)=([j:Z]
   if (within_bound j) then [H1]((at (Z-> (option T)) Vf j H1) i) 
                       else [H2](None T))).
    

(** Axiom of the global synchronisation construction *)
Axiom ifat_def: (T:Set)(V:(Vector bool)) (n:Z)(R_IF,R_ELSE:T)
(Hyp_n: `0 <= n < (bsp_p tt)`) ((ifat T V n R_IF R_ELSE)= 
        (if (sumbool_of_bool (at bool V n Hyp_n)) then [H1] R_IF 
           else [H2] R_ELSE)).
	   
(** Axiom for the parallel superposition *)
Axiom super_def: (T1,T2:Set) (v1: unit -> (Vector T1)) 
  (v2: unit -> (Vector T2)) (super T1 T2 v1 v2)=((v1 tt),(v2 tt)).	

Hints Resolve good_bsp_p mkpar_def apply_def put_def super_def.
Hints Resolve good_bsp_p : zarith.

(** **** Two other axioms. *)

(** Axiom to elim dependent proof for "at", the abstract access to the elements of a parallel vector *)

Axiom at_H: (T:Set; v: (Vector T); i: Z; H1,H2: ` 0<=i<(bsp_p tt)`)
            (at T v i H1)=(at T v i H2).
	    

(**  Axiomatization of the failure of the BSP computer
    - !!! Dangerous, it was only used with "noSome", "hd" and "tl"
    - !!! Do NOT used it !!! *)
Axiom bsp_FAIL:(T:Set) {y:T | True}.
