(* ########################################################################## *)
(* ############################### bsmlbase  ################################ *)
(* ########################################################################## *)

(** Definition of Common BSML expressions, which are offen used in the following. All those functions
are describe in http://bsmllib.free.fr *)

(* ************************************************************************** *)
(* *********************** REPLICATE **************************************** *)
(* ************************************************************************** *)

Require ZArith.
Require Bool.
Require Bool.Sumbool.
Require Omega.
Require PolyList.
Require Bsml.
Require Bsmllemma.
Require Bsmlutils.

Require Export ZArith.
Require Export Bool.
Require Export Bool.Sumbool.
Require Export Omega.
Require Export PolyList.
Require Export Bsml.
Require Export Bsmllemma.
Require Export Bsmlutils.



(** Replication of a value on the parallel vector *)
Definition replicate : (T:Set)(a:T) {res: (Vector T) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T res i Hyp_i)=a}.
Intros.
Exists (mkpar T ([pid:Z]a)).
Intros.
Rewrite mkpar_def;Auto.
Defined.

(** ** Application of a function to parallel vectors *)

(* ************************************************************************** *)
(* ************************** PARFUN **************************************** *)
(* ************************************************************************** *)

(** Application of a function "f" to the elements of a parallel vector *)
Definition parfun : (T1,T2:Set)(f:T1->T2)(v:(Vector T1)) {res: (Vector T2) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T2 res i Hyp_i)= (f (at T1 v i Hyp_i))}.
Intros.
Refine Cases (replicate ? f) of
  (exist p h) => (exist ? ? (apply ? ? p v) ?) end.

Intros.
Rewrite apply_def;Auto.
Rewrite (h i Hyp_i);Trivial.
Defined.

(* ************************************************************************** *)
(* ************************** PARFUN2 *************************************** *)
(* ************************************************************************** *)

(**  Application of a function "f" to the elements of two parallel vectors *)
Definition parfun2 :(T1,T2,T3:Set) (f:T1->T2->T3)(v1:(Vector T1))(v2:(Vector T2)) {res:(Vector T3) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T3 res i Hyp_i)= (f (at T1 v1 i Hyp_i) (at T2 v2 i Hyp_i))}.
Intros.
Refine Cases (parfun ? ? f v1) of
        (exist p h) => (exist ? ? (apply ? ? p v2) ?) end.

Intros.  
Rewrite apply_def;Auto.
Rewrite (h i Hyp_i);Trivial.
Defined.


(* ************************************************************************** *)
(* ************************** PARFUN3 *************************************** *)
(* ************************************************************************** *)

(**  Application of a function "f" to the elements of three parallel vectors *)
Definition parfun3: (T1,T2,T3,T4:Set) (f:T1->T2->T3->T4)(v1:(Vector T1))(v2:(Vector T2))(v3:(Vector T3)) {res:(Vector T4) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T4 res i Hyp_i)= (f (at T1 v1 i Hyp_i) (at T2 v2 i Hyp_i) (at T3 v3 i Hyp_i))}.
Intros.
Refine Cases (parfun2 ? ? ? f v1 v2) of
         (exist p h) => (exist ? ? (apply ? ? p v3) ?) end.

Intros.
Rewrite apply_def;Auto.
Rewrite (h i Hyp_i);Trivial.
Defined.	 


(* ************************************************************************** *)
(* ************************** PARFUN4 *************************************** *)
(* ************************************************************************** *)

(**  Application of a function "f" to the elements of four parallel vectors *)
Definition parfun4: (T1,T2,T3,T4,T5:Set) (f:T1->T2->T3->T4->T5)(v1:(Vector T1))(v2:(Vector T2))(v3:(Vector T3))(v4:(Vector T4)) {res:(Vector T5) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T5 res i Hyp_i)= (f (at T1 v1 i Hyp_i) (at T2 v2 i Hyp_i) (at T3 v3 i Hyp_i) (at T4 v4 i Hyp_i))}.
Intros.
Refine Cases (parfun3 ? ? ? ? f v1 v2 v3) of
         (exist p h) => (exist ? ? (apply ? ? p v4) ?) end.

Intros.
Rewrite apply_def;Auto.
Rewrite (h i Hyp_i);Trivial.
Defined.

(** ** Multiple asynchronous applications *)

(* ************************************************************************** *)
(* *********************** Apply 2   **************************************** *)
(* ************************************************************************** *)

(**  Asynchronous Application for three parallel vectors *)
Definition apply2: (T1,T2,T3:Set)(f:(Vector T1->T2->T3)) (v1:(Vector T1)) (v2:(Vector T2))
  {res:(Vector T3) |  (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T3 res i Hyp_i) = (((at T1->T2->T3 f i Hyp_i) (at T1 v1 i Hyp_i)) (at T2 v2 i Hyp_i))}.
Intros.
Exists (apply ? ? (apply ? ? f v1) v2).
Intros.
Rewrite apply_def; Auto.
Rewrite apply_def; Auto.
Defined.


(* ************************************************************************** *)
(* *********************** Apply 3   **************************************** *)
(* ************************************************************************** *)

(**  Asynchronous Application for four parallel vectors *)
Definition apply3: (T1,T2,T3,T4:Set) (f:(Vector (T1->T2->T3->T4))) (v1:(Vector T1)) (v2:(Vector T2)) (v3:(Vector T3))
 { res: (Vector T4) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T4 res i Hyp_i)=((at ? f i Hyp_i) (at T1 v1 i Hyp_i) (at T2 v2 i Hyp_i) (at T3 v3 i Hyp_i))}.
Intros.
Refine Cases (apply2 ? ? ? f v1 v2) of
  (exist p h) => (exist ? ? (apply ? ? p v3) ?) end.

Intros.
Rewrite apply_def;Auto.
Rewrite (h i Hyp_i);Trivial.
Defined.


(* ************************************************************************** *)
(* *********************** Apply 4   **************************************** *)
(* ************************************************************************** *)

(**  Asynchronous Application for five parallel vectors *)
Definition apply4: (T1,T2,T3,T4,T5:Set) (f:(Vector (T1->T2->T3->T4->T5))) (v1:(Vector T1)) (v2:(Vector T2)) (v3:(Vector T3))(v4:(Vector T4))
 { res: (Vector T5) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at T5 res i Hyp_i)=((at ? f i Hyp_i) (at T1 v1 i Hyp_i) (at T2 v2 i Hyp_i) (at T3 v3 i Hyp_i)(at T4 v4 i Hyp_i))}.
Intros.
Refine Cases (apply3 ? ? ? ? f v1 v2 v3) of
  (exist p h) => (exist ? ? (apply ? ? p v4) ?) end.

Intros.
Rewrite apply_def;Auto.
Rewrite (h i Hyp_i);Trivial.
Defined.

(** ** Some Tools *)

(* ************************************************************************** *)
(* ************************** THIS ****************************************** *)
(* ************************************************************************** *)

(** Create a parallel vector of the pid of the processors *)
Definition this : unit -> {res: (Vector Z) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) (at Z res i Hyp_i)=i}.
Intro.
Exists (mkpar ? ([pid:Z]pid)).
Intros.
Rewrite mkpar_def;Auto.
Defined.

(* ************************************************************************** *)
(* *************************** LAST, NONE, noSome, Hd, Tl ******************* *)
(* ************************************************************************** *)

(** The last pid of process *)
Definition last := [tt:unit] `(bsp_p tt)-1`.

Lemma last_within_bound: `0<=(last tt)<(bsp_p tt)`.
Unfold last.
Generalize good_bsp_p;Intros.
Auto with zarith.
Qed.

Lemma in_last_so_within_bound: (pos:Z) (Hyp_pos:`0 <= pos <= (last tt)`) `0 <= pos < (bsp_p tt)`.
Unfold last.
Intros.
Auto with zarith.
Qed.

Lemma within_bound_so_in_last: (pos:Z)  `0 <= pos < (bsp_p tt)` -> `0 <= pos <= (last tt)`.
Unfold last.
Intros.
Auto with zarith.
Qed.


(** The empty value *)
Definition none := [T1,T2:Set;x:T1] (None T2).

(** To remove the "option" of a value. 
   - !!! raise an exception if the parameter is empty
   - !!! Do NOT used it *)
Definition noSome:(T:Set)(a: (option T))T.
Intros.
Case a;Intros;Auto.
Generalize (bsp_FAIL T);Intro.
Elim H;Intros;Auto.
Defined.


Lemma noSome_irr: (T:Set) (a:(option T)) ~(a=(None T)) -> (a=(Some T (noSome T a))).
Intros.
NewInduction a;Intros;Auto.
Absurd ~(None T)=(None T);Auto.
Qed.


(** Head and tail of a list.
  - !!! raise an exception if the parameter is an empty list
  - !!! Do NOT used them *)
Definition hd:(T:Set)(l:(list T))T.  
Intros.
Case l;Intros;Auto.
Generalize (bsp_FAIL T);Intro.
Elim H;Intros;Auto.
Defined.

Definition tl:(T:Set)(l:(list T)) (list T).
Intros.
Case l;Intros; [ Idtac | Auto].
Generalize (bsp_FAIL (list T));Intro.
Elim H;Intros;Auto.
Defined.


(* ************************************************************************** *)
(* ************************** procs ***************************************** *)
(* ************************************************************************** *)

(** Create a list where the elements of the list are the name of the processes *)
Definition procs: unit -> {res : (list Z) | (Zlength Z res)=(bsp_p tt) /\ ( (i:Z) `0<=i<=(last tt)` -> (Znth_spec Z res `i+1` i))}.
Intro.
Assert `0<=(last tt)`.
Generalize good_bsp_p;Intros.
Unfold last.
Omega.
Refine Cases (from_to `(last tt)` H0) of
    (exist p h) => (exist ? ? p ?) end.
Split.    
Generalize (from_to_Zlength `(last tt)` p h).
Unfold last.
Auto with zarith.
Intros.
Apply (from_to_Nth `(last tt)` H0 p h i H1).
Defined.



(* ************************************************************************** *)
(* ************************ Exchange Vector to pair ************************* *)
(* ************************************************************************** *)

(** Exchange vector of values to pair of values *)
Definition par_of_pair : (T1,T2:Set) (arg:(Vector T1)*(Vector T2)) { res:(Vector (T1*T2)) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) ((Fst (at ? res i Hyp_i))= (at T1 (Fst arg) i Hyp_i)) /\ ((Snd (at ? res i Hyp_i))= (at T2 (Snd arg) i Hyp_i))}.
Intros.
Refine (Cases (replicate ? ([xi:T1][yi:T2] (xi,yi))) of 
   (exist p h) => (Cases (apply2 ? ? ? p (Fst arg) (Snd arg)) of
     (exist p' h') => (exist ? ? p' ?) end)end).

Intros.     
Rewrite (h' i Hyp_i).
Rewrite (h i Hyp_i);Auto.
Defined.


(* ************************************************************************** *)
(* ************************ Exchange Pair to vector ************************* *)
(* ************************************************************************** *)

(** Exchange pair of values to vector of values *)
Definition pair_of_par : (T1,T2:Set) (arg: (Vector T1*T2)) { res : (Vector T1)*(Vector T2) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`)  ( (at T1 (Fst res) i Hyp_i) = (Fst (at ? arg i Hyp_i) )) /\ ( (at T2 (Snd res) i Hyp_i) = (Snd (at ? arg i Hyp_i) ))}.
Intros.
Exists ( (apply ? ? (Cases (replicate ? Fst) of
		            (exist p h) => p end) arg),
		(apply ? ? (Cases (replicate ? Snd) of
		            (exist p h) => p end) arg)).
Intros.	
Split;Simpl;Rewrite apply_def;Auto;Rewrite mkpar_def;Auto.
Defined.



(* ************************************************************************** *)
(* ******************************* MASK  ************************************ *)
(* ************************************************************************** *)

(** The mask composition *)
Definition mask : (T:Set)(c:Z->bool)(x,y:(Vector T)) { res :(Vector T) |(i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`)
 ( ((c i)=true) -> (at T res i Hyp_i)=(at T x i Hyp_i) ) /\ ( ((c i)=false) -> (at T res i Hyp_i)=(at T y i Hyp_i))}.
Intros.
Exists  (apply T T (apply T (T->T) (mkpar (T->T->T) [j:Z]
 if (sumbool_of_bool (c j)) then [H1] ([x:T][y:T]x) else [H2]([x:T][y:T]y)) x) y).

Intros.
Split;Intros;Rewrite apply_def;Auto;Rewrite apply_def;Auto;Rewrite mkpar_def;Auto;Rewrite H;Elim sumbool_of_bool;Intros;Auto;Discriminate.
Defined.



(* ************************************************************************** *)
(* ****************************  noSome_vect  ****************************** *)
(* ************************************************************************** *)

(** Remove the "option" of a parallel vector if and only if there is no empty value in the parallel vector *)
Definition noSome_vect: (T:Set) (v:(Vector (option T)))
( (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)  ~( (at (option T) v i Hyp_i) = (None T) ))  -> 
 { res:(Vector T) | (i:Z) (Hyp_i: `0<=i<(bsp_p tt)`)  (Some T (at T res i Hyp_i))=(at (option T) v i Hyp_i)}.

 Intros.
 Refine Cases (parfun (option T) ? (noSome T) v) of
    (exist p h) => (exist ? ? p ?) end.

 Intros.
 Generalize (H i Hyp_i);Intro.
 Generalize (no_None_so_Some T (at (option T) v i Hyp_i) H0);Intro.
 Inversion H1.
 Generalize (h i Hyp_i).
 Rewrite H2.
 Unfold noSome;Intro.
 Rewrite H3;Auto.
 Defined.

(* ************************************************************************** *)
(* ****************************  tl_hd_vect  ******************************** *)
(* ************************************************************************** *)

(** Give the head and the tail of a parallel vector of list if and only if there is no empty list in the parallel vector *)
Definition tl_hd_vect: (T:Set) (v:(Vector (list T)))
( (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)  ~( (at (list T) v i Hyp_i) = (nil T) ))  -> 
 { res:(Vector T*(list T)) | (i:Z) (Hyp_i: `0<=i<(bsp_p tt)`)  (cons (Fst (at ? res i Hyp_i)) (Snd (at ? res i Hyp_i)))=(at (list T) v i Hyp_i)}.

 Intros.
 Refine Cases (parfun (list T) ? ([l:(list T)] ((hd T l),(tl T l))) v) of
    (exist p h) => (exist ? ? p ?) end.

 Intros.
 Generalize (H i Hyp_i);Intro.
 Generalize (no_nil_so_hd_tl T (at (list T) v i Hyp_i) H0);Intro.
 Inversion H1.
 Inversion H2.
 Generalize (h i Hyp_i).
 Rewrite H3.
 Simpl;Intros.
 Rewrite H4;Auto.
 Defined.


(* ************************************************************************** *)
(* ****************************  no_some_in_list_vect  ********************** *)
(* ************************************************************************** *)

Definition no_some_in_list_vect : (T:Set) (v:(Vector (list (option T))))
( (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`) 
    ( (`(Zlength (option T) (at (list (option T)) v i Hyp_i))=0`) \/
  ((`1<=(Zlength (option T) (at (list (option T)) v i Hyp_i))` -> (n:Z) (Hyp_n:(`1<=n<=(Zlength (option T) (at (list (option T)) v i Hyp_i))`))
  (val:(option T)) (Znth_spec (option T) (at (list (option T)) v i Hyp_i) n val) -> ~(val = (None T)))))) -> 
{res : (Vector (list T)) | (i:Z) (Hyp_i: `0<=i<(bsp_p tt)`) 
                             (  ( (Zlength (option T)  (at ? v i Hyp_i))=(Zlength T (at ? res i Hyp_i))) /\
			     (`1<=(Zlength (option T) (at (list (option T)) v i Hyp_i))` ->
                           (n:Z) (Hyp_n:(`1<=n<=(Zlength T (at (list T) res i Hyp_i))`))
			   (val2:(option T)) (Znth_spec (option T) (at (list (option T)) v i Hyp_i) n val2)
			    -> val2=(Some T (Znth T (at (list T) res i Hyp_i) n Hyp_n))))}.

Intros.
Refine Cases (parfun ? ? ( (map 1!(option T) 2!T) (noSome T) ) v) of
    (exist p h) => (exist ? ? p ?) end.

Intros.
Rewrite (h i Hyp_i).
Split;Intros.
Generalize (Zlength_map (option T) T (at (list (option T)) v i Hyp_i) (noSome T));Auto.

Generalize (H i Hyp_i);Intro.
Clear H h.
Elim H2;Clear H2;Intros.
Absurd  `(Zlength (option T) (at (list (option T)) v i Hyp_i)) <> 0`;Auto.
Apply bidon5;Auto.

Assert `1 <= n <= (Zlength (option T) (at (list (option T)) v i Hyp_i))`.
Assert ( (Zlength (option T)  (at ? v i Hyp_i)) = (Zlength T (map (noSome T) (at (list (option T)) v i Hyp_i)) )).
Generalize (Zlength_map (option T) T (at (list (option T)) v i Hyp_i) (noSome T));Auto.
Generalize Hyp_n.
Rewrite (egal_reflexive ? ? ? H2);Auto.

Generalize ( H H0 n H2 val2 H1);Intro.
Clear H.
Rewrite (Znth_ind (option T) T (at (list (option T)) v i Hyp_i) (noSome T) n Hyp_n val2 H1).
Apply noSome_irr;Auto.
Defined.

 
(* ************************************************************************** *)
(* ****************************  Get_one  *********************************** *)
(* ************************************************************************** *)

(** ** Certified implementation of the "get" operator
     the get operator is synchrone exchange of values the BS"lambda"-calculus *)

Definition get_one_tmp : (T:Set)(datas:(Vector T)) (pids:(Vector Z)) 
(Hyp_pids: (pid:Z) (Hyp_pid: `0<=pid<(bsp_p tt)`) `0<=(at Z pids pid Hyp_pid)<(bsp_p tt)`)
 {res : (Vector (option T)) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
    (at ? res i Hyp_i)= (Some T (at T datas (at Z pids i Hyp_i) (Hyp_pids i Hyp_i)))}.

Intros.
Refine  let replace_by_data = (parfun2 ? ? ? ([f:?][d:T][dst:Z] Cases (f dst) of 
                                                                     (Some tt) => (Some T d) 
                                                                    | None => (None T) end)) in
 Cases (parfun ? ? ([i,dst:Z] if (Z_eq_dec dst i) then ([H1](Some unit tt)) else ([H2](None unit))) pids) of
         (exist p1 h1) =>  let ask=(put ?  p1) in 
          (Cases (replace_by_data ask datas) of
              (exist p2 h2) => let reply=(put ? p2) in  (exist ? ?  (apply ? ? reply pids) ?) end) end.

Intros.
Rewrite apply_def;Auto.
Unfold reply.
Rewrite put_def;Auto.
Case (within_bound (at Z pids i Hyp_i));Intro.
Rewrite (h2 (at Z pids i Hyp_i) a).
Unfold ask.
Rewrite put_def;Auto.
Elim (within_bound i);Intros;Auto.
Rewrite (h1 i a0).
Elim (Z_eq_dec (at Z pids i Hyp_i) (at Z pids i a0));Intros;Auto.
Rewrite (at_H T datas (at Z pids i Hyp_i) a (Hyp_pids i Hyp_i));Auto.
Assert `(at Z pids i Hyp_i) <> (at Z pids i Hyp_i)`.
Generalize (at_H Z pids i Hyp_i a0);Intro.
Rewrite H;Auto.
Absurd `(at Z pids i Hyp_i) <> (at Z pids i Hyp_i)`;Auto.
Absurd ~`0 <= i < (bsp_p tt)`;Auto.
Absurd  ~`0 <= (at Z pids i Hyp_i) < (bsp_p tt)`;Auto.
Defined.


Definition get_one : (T:Set)(datas:(Vector T)) (pids:(Vector Z)) 
(Hyp_pids: (pid:Z) (Hyp_pid: `0<=pid<(bsp_p tt)`) `0<=(at Z pids pid Hyp_pid)<(bsp_p tt)`)
 {res : (Vector T) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`) 
    (at ? res i Hyp_i)= (at T datas (at Z pids i Hyp_i) (Hyp_pids i Hyp_i))}.
     
Intros.
Refine Cases (get_one_tmp T datas pids Hyp_pids) of
	 (exist p h) => Cases (noSome_vect ? p ?) of
	                  (exist p' h') => (exist ? ? p' ?) end end;Intros.

Rewrite (h i Hyp_i).
Discriminate.
Apply egal_some.
Rewrite (h' i Hyp_i).
Rewrite (h i Hyp_i);Auto.
Defined.

(* ************************************************************************** *)
(* ****************************  Get_list  ********************************** *)
(* ************************************************************************** *)

(** ** Certified implementation of the "get_list" operator *)

Lemma get_list_trans : (l:(list Z)) ( (l=(nil Z)) \/ ( (`1<=(Zlength Z l)`) -> 
  (n:Z) `1<=n<=(Zlength Z l)` -> (pid:Z) (Znth_spec Z l n pid) 
  -> `0<=pid<(bsp_p tt)`)) -> (`1<=(Zlength Z l)`) -> (n:Z) `1<=n<=(Zlength Z l)` 
  -> (pid:Z) (Znth_spec Z l n pid) -> `0<=pid<(bsp_p tt)`.
Intros.
Elim H;Simpl;Intro;Auto.
Generalize H0;Rewrite H3;Simpl;Intro.
Absurd `1 <= 0`;Auto with zarith.
Apply (H3 H0 n);Auto.
Qed.

Definition get_list_tmp: (T:Set) (datas:(Vector T)) (lpids:(Vector (list Z)))
(Hyp_lpids: (i:Z) (Hyp_i: `0<=i<(bsp_p tt)`)
 (  ( (at (list Z) lpids i Hyp_i)=(nil Z)) \/
  ( (`1<=(Zlength Z (at (list Z) lpids i Hyp_i))`) -> 
    (n:Z) (H_l:(`1<=n<=(Zlength Z (at (list Z) lpids i Hyp_i))`))
    (pid:Z) (Znth_spec Z (at ? lpids i Hyp_i) n pid) -> `0<= pid <(bsp_p tt)`)))

 {res : (Vector (list (option T))) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
    ( ( (Zlength Z  (at ? lpids i Hyp_i))=(Zlength (option T) (at ? res i Hyp_i))) /\
        ((H_l:(`1<=(Zlength Z (at (list Z) lpids i Hyp_i))`)) (n:Z) (Hyp_n:(`1<=n<=(Zlength Z (at (list Z) lpids i Hyp_i))`))
        (pid:Z) (H_pid:(Znth_spec Z (at (list Z) lpids i Hyp_i) n pid))
        (Znth_spec (option T) (at (list (option T)) res i Hyp_i) n (Some T (at T datas pid (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid))))))}.
 
Intros.
Refine  
  Cases (parfun ? ? ( [l:(list Z);dst:Z]  if (In_dec Z_eq_dec dst l) then [H1] (Some unit tt)
                                              else [H2](None unit)) lpids) of
     (exist p h) => let ask=(put ? p) in
   let replace_by_data= (parfun2 ? ? ? ( [f:(Z->(option unit)); d:T; dst:Z] Cases (f dst) of
                                                                                       (Some tt) => (Some T d)
										     |  None     => (None T) end)) in
     Cases (replace_by_data ask datas) of
        (exist p' h') => let reply=(put ? p') in
       Cases (parfun2 ? ? ? (map 1!Z 2!(option T)) reply lpids) of
         (exist p'' h'') => (exist ? ? p'' ?) end end end.
       
Intros.    
Rewrite (h'' i Hyp_i).
Clear h''.
Unfold reply.
Split;Intros.
Generalize (Zlength_map Z (option T) (at (list Z) lpids i Hyp_i) (at (Z->(option T)) (put T p') i Hyp_i));Auto.

Apply (map_Znth2  Z (option T)(at (list Z) lpids i Hyp_i)  
         (at Z->(option T) (put T p') i Hyp_i) pid  (Some T (at T datas pid (get_list_trans (at (list Z) lpids i Hyp_i)
           (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid) )));Auto.
Rewrite put_def.
Case (within_bound pid);Intros.
Rewrite (h' pid a).
Clear h' replace_by_data.
Unfold ask.
Clear ask.
Rewrite put_def.
Case  (within_bound i);Intros.
Rewrite (h i a0).
Clear h.
Case (In_dec Z_eq_dec pid (at (list Z) lpids i a0));Intros.
Apply some_egal.
Apply at_H.
Assert (at (list Z) lpids i Hyp_i)= (at (list Z) lpids i a0).
Apply at_H.
NewInduction H.
Generalize (if_Znth_then_In Z (at (list Z) lpids i Hyp_i) n Hyp_n pid H_pid);Intro.

Absurd ~(In pid (at (list Z) lpids i Hyp_i));Auto.
Absurd ~`0 <= i < (bsp_p tt)`;Auto.

Clear replace_by_data h' h.
Generalize (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid);Intro.
Absurd  ~`0 <= pid < (bsp_p tt)`;Auto.
Defined.


Definition get_list: (T:Set) (datas:(Vector T)) (lpids:(Vector (list Z)))
 (Hyp_lpids: (i:Z) (Hyp_i: `0<=i<(bsp_p tt)`)
 (  ( (at (list Z) lpids i Hyp_i)=(nil Z)) \/
  ( (`1<=(Zlength Z (at (list Z) lpids i Hyp_i))`) -> 
    (n:Z) (H_l:(`1<=n<=(Zlength Z (at (list Z) lpids i Hyp_i))`))
    (pid:Z) (Znth_spec Z (at ? lpids i Hyp_i) n pid) -> `0<= pid <(bsp_p tt)`)))
 {res : (Vector (list T)) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
      ( ( (Zlength Z  (at ? lpids i Hyp_i))=(Zlength T (at ? res i Hyp_i))) /\
        ((H_l:(`1<=(Zlength Z (at (list Z) lpids i Hyp_i))`)) (n:Z) (Hyp_n:(`1<=n<=(Zlength Z (at (list Z) lpids i Hyp_i))`))
        (pid:Z) (H_pid:(Znth_spec Z (at (list Z) lpids i Hyp_i) n pid))
        (Znth_spec T (at (list T) res i Hyp_i) n (at T datas pid (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid)))))}.

Intros.
Refine Cases (get_list_tmp T datas lpids Hyp_lpids) of
	 (exist p h) => Cases (no_some_in_list_vect ? p ?) of
	                  (exist p' h') => (exist ? ? p' ?) end end;Intros.

Generalize (h i Hyp_i);Intro.
Clear h.
Inversion H.
Clear H.
Generalize (Hyp_lpids i Hyp_i);Intro.
Elim H;Intros.
Left.
Rewrite (egal_reflexive ? ? ? H0).
Rewrite H2;Simpl;Auto.

Right;Intros.
Generalize H3.
Rewrite (egal_reflexive ? ? ? H0);Intro.
Generalize Hyp_n.
Rewrite (egal_reflexive ? ? ? H0);Intro.
Generalize (H1 H5 n Hyp_n0);Intro.
Generalize (Znth_exist Z (at (list Z) lpids i Hyp_i) n Hyp_n0);Intro.
Inversion H7.
Generalize (H6 x H8);Intro.
Assert val=(Some T (at T datas x (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H5 n Hyp_n0 x H8))).
Apply (Znth_egal (option T) (at (list (option T)) p i Hyp_i) n val);Auto.
Rewrite H10.
Discriminate.

Split;Intros.
Generalize (h i Hyp_i);Intro.
Inversion H.
Generalize (h' i Hyp_i);Intro.
Inversion H2.
Clear H4 H2 H1.
Rewrite H0.
Rewrite H3;Reflexivity.

Generalize (h i Hyp_i);Intro.
Clear h.
Inversion H.
Clear H.
Generalize (h' i Hyp_i);Intro.
Clear h'.
Inversion H.
Clear H.
Assert  `1 <= (Zlength (option T) (at (list (option T)) p i Hyp_i))`.
Rewrite (egal_reflexive  ? ? ? H0);Auto.
Assert `1 <= n <= (Zlength T (at (list T) p' i Hyp_i))`.
Rewrite (egal_reflexive  ? ? ? H2).
Rewrite (egal_reflexive  ? ? ? H0);Auto.
Generalize (H1 H_l n Hyp_n pid H_pid);Intro.
Clear H1.
Generalize (H3 H n H4 (Some T (at T datas pid (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid))) H5);Intro.
Assert (at T datas pid (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid))=(Znth T (at (list T) p' i Hyp_i) n H4).
Apply egal_some;Auto.
Apply (Znth_Znth_spec T (at (list T) p' i Hyp_i)(at T datas pid (get_list_trans (at (list Z) lpids i Hyp_i) (Hyp_lpids i Hyp_i) H_l n Hyp_n pid H_pid)) n H4);Auto.
Defined.



(* ************************************************************************** *)
(* ****************************  bsml_print  ******************************** *)
(* ************************************************************************** *)

(** ** Certified implementation of the "bsml_print" and "parprint" operators *)

Definition bsml_print: (T:Set) (p:(T -> unit)) (at_pid:Z) (H_atpid:`0<=at_pid<(bsp_p tt)`) (vec:(Vector T)) 
{res :(Vector unit) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`) 
 ( (i=at_pid) -> (at unit res i Hyp_i)=(p (at T vec i Hyp_i))) /\ ( (`i<>at_pid`) -> (at unit res i Hyp_i)=tt)}.
Intros.
Refine let nothing = ([x:T] tt) in 
       let print=(apply ? ? (mkpar ? ([pid:Z] if (Z_eq_dec pid at_pid) then [H1] p else [H2] nothing)) vec) in
         if (within_bound at_pid) then [H1] (exist ? ? print ?)
	   else [H2] ?;Intros.

Split;Unfold print;Rewrite apply_def;Rewrite mkpar_def;Case (Z_eq_dec i at_pid);Intros;Auto.
Absurd `i = at_pid`;Auto with zarith.
Absurd `i <> at_pid`;Auto with zarith.
Elim H2;Auto.
Defined.


Definition parprint: (T:Set) (p:(T -> unit))  (vec:(Vector T)) {res :(Vector unit) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`) (at unit res i Hyp_i)=(p (at T vec i Hyp_i))}.
Intros.
Refine Cases (this tt) of
  (exist pThis hThis) => Cases (parfun2 ? ? ? ([pid:Z;data:T] (p data)) pThis vec) of
    (exist pRes hRes) => (exist ? ? pRes ?) end end;Intros.

Apply hRes;Auto.
Defined.


(* ************************************************************************** *)
(* ****************************  super_tools  ******************************* *)
(* ************************************************************************** *)
(** ** Tools for the parallel superposition *)


Definition inbounds : (first,last,n:Z) {`first<=n<=last`}+{~`first<=n<=last`}.
Intros.
Case (Z_le_dec first n);Intros.
Case (Z_le_dec n last0);Intros;Auto.
Intuition.
Intuition.
Defined.

Definition mix : (T:Set) (m:Z) (v:(Vector T)*(Vector T))
 {res : (Vector T) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`) 
  ( (`i<=m`) -> (at ? res i Hyp_i)=(at ? (Fst v) i Hyp_i)) 
   /\
   ( (~`i<=m`) -> (at ? res i Hyp_i)=(at ? (Snd v) i Hyp_i))}.
Intros.
Refine let f = ([pid:Z;v1,v2:T] if (Z_le_dec pid m) then [H1] v1 else [H2] v2) in
 (exist ? ? (apply ? ? (apply ? ? (mkpar ? f) (Fst v)) (Snd v)) ?).
Intros.
Split;Intro;Rewrite apply_def;Auto;Rewrite apply_def;Auto;Rewrite mkpar_def;Auto;Unfold f;Case (Z_le_dec i m);Intros;Auto.
Absurd ~`i <= m`;Auto.
Absurd  `i <= m`;Auto.
Defined.
