(* ########################################################################## *)
(* ############################### Bsmlcomm  ################################ *)
(* ########################################################################## *)

(** This module describe communication primitives. All 
    those function are describe in http://bsmllib.free.fr *)

Require ZArith.
Require Coq.ZArith.Zdiv.
Require Omega.
Require PolyList.
Require Bsml.
Require Bsmlbases.
Require Bsmllemma.

Require Export ZArith.
Require Export Coq.ZArith.Zdiv.
Require Export Omega.
Require Export PolyList.
Require Export Bsml.
Require Export Bsmlbases.
Require Export Bsmllemma.


(* ************************************************************************** *)
(* ************************** SHIFT ***************************************** *)
(* ************************************************************************** *)


(** ** Shift values *)

Definition shift_tmp : (T:Set)(dec:Z)(datas:(Vector T)) { res:(Vector (option T)) | (i:Z) (H_i:`0<=i<(bsp_p tt)`)
 (at ? res i H_i)=(Some ? (at ? datas `((bsp_p tt)+i-(dec%(bsp_p tt)))% (bsp_p tt)` (shift_dec dec i H_i)))}.
Intros.
Refine 
    Cases (this tt) of
      (exist p h) => Cases (parfun2 ? ? ? ( [data:T;pid,dst:Z] (Some ? data)) datas p) of
          (exist p' h') => ( exist ? ? (apply ? ? (put ? p') (mkpar ? ([pid:Z] `((bsp_p tt)+pid-(dec%(bsp_p tt)))%(bsp_p tt)`))) ?)
	    end
	end.
Intros.	
Rewrite apply_def;Auto.
Rewrite mkpar_def;Auto.
Rewrite put_def;Auto.
Case (within_bound  `((bsp_p tt)+i-dec%(bsp_p tt))%(bsp_p tt)`);Intros.
Rewrite (h' `((bsp_p tt)+i-dec%(bsp_p tt))%(bsp_p tt)` a).
Apply some_egal.
Rewrite (at_H T datas `((bsp_p tt)+i-dec%(bsp_p tt))%(bsp_p tt)` a  (shift_dec dec i H_i));Trivial.
Elim n.
Apply (shift_dec  dec i H_i).
Defined.


(* ************************** SHIFT ***************************************** *)
(** *** General Shift *)
Definition shift: (T:Set)(dec:Z)(datas:(Vector T)) { res:(Vector T) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
 (at ? res i Hyp_i)=(at ? datas `((bsp_p tt)+i-(dec%(bsp_p tt)))% (bsp_p tt)` (shift_dec dec i Hyp_i))}.
Intros. 
Refine Cases (shift_tmp T dec datas) of
	 (exist p h) => Cases (noSome_vect ? p ?) of
	                  (exist p' h') => (exist ? ? p' ?) end end;Auto;Intros.

Rewrite (h i Hyp_i).
Discriminate.
Apply egal_some.
Rewrite (h' i Hyp_i).
Rewrite (h i Hyp_i);Auto.
Defined.


(* ************************** shift_left ************************************ *)
(** *** Shift left*)

Definition shift_left : (T:Set)(data:(Vector T))
{res : (Vector T) | (i:Z) (Hyp_i:`0 <= i < (bsp_p tt)`)

(`i=0` -> (at ? res i Hyp_i)=(at T data (last tt) last_within_bound)) /\
   ( (Hyp_i2:`i>0`) (at ? res i Hyp_i)=(at ? data `i-1` (p_sub i Hyp_i Hyp_i2)))}.
Intros. 
Refine Cases (shift T `1` data) of
	  (exist p h) => (exist ? ? p ?) end.
Split;Intros.
Rewrite (h i Hyp_i).
Unfold last.
Apply (at_egal T  `((bsp_p tt)+i-1%(bsp_p tt))%(bsp_p tt)`  (shift_dec `1` `i` Hyp_i) `(bsp_p tt)-1` last_within_bound).
Rewrite H.
Apply shift_left1.

Rewrite (h i Hyp_i).
Apply (at_egal T  `((bsp_p tt)+i-1%(bsp_p tt))%(bsp_p tt)` (shift_dec `1` i  Hyp_i) `i-1` (p_sub i Hyp_i Hyp_i2)).
Assert `0 < i < (bsp_p tt)`.
Auto with zarith.
Apply shift_left2;Auto.
Defined.


(* ************************** shift_right *********************************** *)
(** *** Shift right *)

Lemma shift_right_dec : (i:Z) `0<= i < (bsp_p tt)` -> `i< (last tt)` -> `0<=i+1<(bsp_p tt)`.
Unfold last.
Intros.
Auto with zarith.
Qed.

Axiom shift_right2: (i:Z)  `0 <= i < (last tt)` ->  `((bsp_p tt)+i-(-1)%(bsp_p tt))%(bsp_p tt) = i+1`.

Definition shift_right : (T:Set)(data:(Vector T))
{res : (Vector T) |(i:Z) (Hyp_i:`0<= i < (bsp_p tt)` )
 ( `i=(last tt)` -> (at ? res i Hyp_i)=(at T data `0` zero_within_bound)) /\
   ( (Hyp_i2:`i<(last tt)`) ((at ? res i Hyp_i)=(at ? data `i+1` (shift_right_dec i Hyp_i Hyp_i2))))}.
Intros.     
Refine Cases (shift T `-1` data) of
	  (exist p h) => (exist ? ? p ?) end.

Split;Intros.
Rewrite (h i Hyp_i).
Apply (at_egal T `((bsp_p tt)+i-(-1)%(bsp_p tt))%(bsp_p tt)` (shift_dec `-1` i Hyp_i) `0` zero_within_bound).
Rewrite H.
Unfold last.
Apply shift_right1.
Rewrite (h  i Hyp_i).
Apply (at_egal T  `((bsp_p tt)+i-(-1)%(bsp_p tt))%(bsp_p tt)`(shift_dec `-1` i Hyp_i) `i+1` (shift_right_dec i Hyp_i Hyp_i2)).
Assert `0 <= i < (last tt)`.
Unfold last.
Auto with zarith.
Apply shift_right2;Auto.
Defined.



(* ************************************************************************** *)
(* *********************** Total Exchange *********************************** *)
(* ************************************************************************** *)

(** ** Functions of total exchanges *)

Definition totex_tmp : (T:Set; vec:(Vector T)) {res:(Vector (Z->(option T))) | (i:Z) (Hyp_i:`0 <= i < (bsp_p tt)`)
 (src:Z) ( (Hyp_src:`0<=src<(bsp_p tt)`)  ((at (Z->(option T)) res i Hyp_i) src)=( Some T (at T vec src Hyp_src)))
/\  ( (Hyp_src:~`0<=src<(bsp_p tt)`) ((at (Z->(option T)) res i Hyp_i) src) = (None T))}.

Intros.
Refine Cases (parfun ? ? ([data:T;dst:Z] (Some T data)) vec) of
         (exist p h') => (exist ? ? (put ? p) ?) end.

Split;Intros;Rewrite put_def;Auto;Case (within_bound src);Intros;Auto.
Rewrite (h' src a).
Apply some_egal.
Apply (at_H T vec src a Hyp_src);Auto.
Absurd ~`0 <= src < (bsp_p tt)`;Auto.
Absurd ~`0 <= src < (bsp_p tt)`;Auto.
Defined.

(** *** Totex *)
Definition totex : (T:Set; vec:(Vector T)) {res:(Vector Z->T) | (i:Z) (Hyp_i:`0 <= i < (bsp_p tt)`)
 (src:Z) (Hyp_src:`0<=src<(bsp_p tt)`)  ((at ? res i Hyp_i) src)=(at T vec src Hyp_src)}.
Intros.

Refine Cases (totex_tmp T vec ) of
    (exist p h) => Cases (parfun ? ? ([f:?;src:Z] (noSome ? (f src))) p) of
      (exist p' h') => (exist ? ? p' ?) end end.
Intros.
Rewrite (h' i Hyp_i).
Generalize (h i Hyp_i src);Intros.
Inversion H.
Rewrite (H0 Hyp_src);Simpl;Trivial.
Defined.


(** *** Total exchange *)

Definition total_exchange : (T:Set; vec:(Vector T)) {res:(Vector (list T)) | (i:Z) (Hyp_i:`0 <= i < (bsp_p tt)`)
`(Zlength T (at (list T) res i Hyp_i))=(bsp_p tt)` /\
(pos:Z) (Hyp_pos:`0 <= pos < (bsp_p tt)`) (Znth_spec T (at (list T) res i Hyp_i) `pos+1` (at T vec pos Hyp_pos))}.
Intros.
Refine Cases (procs tt) of
   (exist lpar hpar) => Cases (replicate ? lpar) of
     (exist vec_par hvec_par) => Cases (totex T vec) of 
       (exist ptot htot) => Cases (parfun2 (Z->T) (list Z) (list T) (map 1!Z 2!T) ptot vec_par) of
        (exist p' h') => (exist ? ? p' ?) end end end end.

Intros.
Rewrite (h' i Hyp_i).
Rewrite (hvec_par i Hyp_i).
Inversion hpar.
Split;Intros.
Rewrite (Zlength_map Z T lpar (at (Z->T) ptot i Hyp_i));Trivial.
Assert (Znth_spec Z lpar `pos+1` pos).
Apply H0.
Apply within_bound_so_in_last;Auto.
Assert ((at T vec pos Hyp_pos)=((at Z->T ptot i Hyp_i) pos)).
Rewrite (htot i Hyp_i pos Hyp_pos);Auto.
Apply (map_Znth2 Z T lpar  (at Z->T ptot i Hyp_i) pos (at T vec pos Hyp_pos) H2 `pos+1` H1).
Defined.



(* ************************************************************************** *)
(* *********************** Bcast_direct ************************************* *)
(* ************************************************************************** *)

(** ** Broadcasting values
   We give here some algorithms of broadcasting. *)

(** ***  Direct broadcast *)
Definition  bcast_direct_tmp: (T:Set)(root:Z) (Hyp_root:`0<= root < (bsp_p tt)`) (vv:(Vector T))
  {res:(Vector (option T)) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) ((at (option T) res i Hyp_i)=(Some T (at T vv root Hyp_root))) }.
Intros.
Refine let tosend=(mkpar (T->Z->(option T)) ([pid:Z][v:T] if (Z_eq_dec pid root) then [H1] ([dst:Z](Some T v)) 
                                                            else [H2]([dst:Z](None T)) ))
	in let recv=(put T (apply T (Z->(option T)) tosend vv))
	in Cases (replicate Z root) of
	  (exist p h') => (exist ? ? (apply Z (option T) recv p) ?)
	      end.
Intros.
Rewrite apply_def; Auto.
Unfold recv.
Rewrite put_def;Auto.
Case (within_bound (at Z p i Hyp_i));Intros;Auto.
Rewrite apply_def;Auto.
Unfold tosend.
Rewrite mkpar_def;Auto.
Case (Z_eq_dec (at Z p i Hyp_i) root) ;Intros.
Generalize a; Rewrite e; Intros.
Rewrite (at_H T vv root a0 Hyp_root);Auto.
Absurd `(at Z p i Hyp_i) <> root`;Auto.
Elim n;Rewrite (h' i Hyp_i);Trivial.
Defined.


Definition bcast_direct :(T:Set)(root:Z)(vv:(Vector T)) (Hyp_root:`0<= root <  (bsp_p tt)`)
  {res:(Vector  T) | (i:Z) (Hyp_i:`0 <=i < (bsp_p tt)`) ((at T res i Hyp_i)=(at T vv root Hyp_root)) }.
Intros.
Refine Cases (bcast_direct_tmp ? root Hyp_root vv) of
	 (exist p h) => Cases (noSome_vect ? p ?) of
	                  (exist p' h') => 
    if (within_bound root) then [H1] (exist ? ? p' ?) 
      else [H2] ? end end;Auto;Intros.

Rewrite (h i Hyp_i).
Discriminate.
Apply egal_some.
Rewrite (h' i Hyp_i).
Rewrite (h i Hyp_i);Auto.
Elim H2;Auto.
Defined.

(** *** totex broadcast *)

(** *** 2 phases algorithms *)

(* ************************************************************************** *)
(* ************************** Scatter *************************************** *)
(* ************************************************************************** *)

(** ** Scatter values of a parallel vector *)

(** *** Simply Scatter values *)
Definition scatter_tmp : (T1,T2:Set)(partition:(T1->Z->(option T2)))(root:Z)
                         (Hyp_root:`0<=root<(bsp_p tt)`) (v:(Vector T1))
(Hyp:((i:Z) `0<=i<(bsp_p tt)` -> ~(partition (at T1 v root Hyp_root) i)=(None T2)))
 { res :(Vector (option T2)) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
                 (at ? res i Hyp_i)= (partition (at T1 v root Hyp_root) i)}.
      
Intros.
Refine let mkmsg=(mkpar ? [pid:Z] if (Z_eq_dec pid root) then ([H1]partition) 
                                    else [H2]([a:T1](none Z T2))) in
        let msg=(put ? (apply ? ? mkmsg v)) in
	 Cases (replicate ? root) of
	  (exist p1 h1) =>(exist ? ? (apply ? ? msg p1) ? )
	    end.

Intros. 
Rewrite apply_def;Auto.
Unfold msg.
Rewrite put_def;Auto.
Case (within_bound (at Z p1 i Hyp_i));Intros.     
Rewrite apply_def;Auto.
Unfold mkmsg.
Rewrite mkpar_def;Auto.
Case (Z_eq_dec (at Z p1 i Hyp_i) root);Intros.
Generalize a; Rewrite e; Intros.
Rewrite (at_H T1 v root Hyp_root a0);Auto.
Unfold none.
Generalize (h1 i Hyp_i);Intros.
Absurd `(at Z p1 i Hyp_i) <> root`;Auto.
Generalize (h1 i Hyp_i);Intros.
Generalize Hyp_root.
Rewrite (egal_reflexive ? ? ? H);Intro.
Absurd  ~`0 <= (at Z p1 i Hyp_i) < (bsp_p tt)`;Auto.
Defined.


Definition scatter : (T1,T2:Set)(partition:(T1->Z->(option T2)))(root:Z)
                         (Hyp_root:`0<=root<(bsp_p tt)`) (v:(Vector T1))
(Hyp:((i:Z) `0<=i<(bsp_p tt)` -> ~(partition (at T1 v root Hyp_root) i)=(None T2)))
 { res :(Vector T2) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
                 (Some ? (at ? res i Hyp_i))= (partition (at T1 v root Hyp_root) i)}.

Intros.
Refine Cases (scatter_tmp T1 T2 partition root Hyp_root v Hyp) of
	 (exist p h) => Cases (noSome_vect ? p ?) of
	                  (exist p' h') => 
    if (within_bound root) then [H1] (exist ? ? p' ?) 
      else [H2] ? end end;Auto;Intros.

Rewrite (h i Hyp_i);Auto.
Rewrite (h' i Hyp_i).
Rewrite (h i Hyp_i);Auto.
Elim H2;Auto.
Defined.

(** *** Scatter a parallel vector of list  *)

(** ** Gather values of a parallel vector *)

(** *** Simple gather *)

Definition gather: (T:Set) (root:Z) (vv:(Vector T)) 
 { res:(Vector (Z->(option T))) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)

 ( (`i=root`)-> (dest:Z) ( ((H_dest:`0<=dest<(bsp_p tt)`)
       ((at ? res i Hyp_i) dest)=(Some ? (at ? vv dest H_dest))) 
             /\
       ((~`0<=dest<(bsp_p tt)`) -> ((at ? res i Hyp_i) dest)=(None ?))))
  /\
 ((`i<>root`) -> (dest:Z) ((at ? res i Hyp_i) dest)=(None ?))}.


Intros.
Refine let mkmsg = [pid:Z] [v:T] [dest:Z] if (Z_eq_dec dest root) then [H1] (Some T v) 
                                 else [H2] (None T) in
        (exist ? ? (put ? (apply ? ? (mkpar ? mkmsg) vv)) ?).				 
Intros.	
Split;Intros;Rewrite put_def;Auto.
Split;Intros; Case (within_bound dest);Intros;Auto.

Rewrite apply_def;Auto.
Rewrite mkpar_def;Auto.
Compute.
Case (Z_eq_dec i root);Intros.
Rewrite (at_H T vv dest a H_dest);Auto.
Absurd `i <> root`;Auto.
Absurd ~`0 <= dest < (bsp_p tt)`;Auto.

Rewrite apply_def;Auto.
Rewrite mkpar_def;Auto.
Compute.
Case (Z_eq_dec i root);Intros;Auto.
Absurd  ~`0 <= dest < (bsp_p tt)`;Auto.

Case (within_bound dest);Intros;Auto.
Rewrite apply_def;Auto.
Rewrite mkpar_def;Auto.
Compute.
Case (Z_eq_dec i root);Intros;Auto.
Absurd `i = root`;Auto.
Defined.

(** *** Gather a list of parallel vector *)

Definition procs_lists_gather : (root:Z) {res : (Vector (list Z)) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
(i=root -> ( ((Zlength Z (at ? res i Hyp_i))=(bsp_p tt)) /\ (pos:Z) `0<=pos<=(bsp_p tt)-1` 
-> (Znth_spec Z (at ? res i Hyp_i) `pos+1` pos))) /\ (`i<>root` -> (at ? res i Hyp_i)=(nil Z))}.
Intros.
Exists (mkpar ? ([pid:Z] if (Z_eq_dec root pid) then [H1](from_to2 `(bsp_p tt)-1`) else [H2] (nil Z))).
Intros.
Assert  `0 <=(bsp_p tt)-1`.
Auto with zarith.
Rewrite (mkpar_def);Auto.
Case (Z_eq_dec root i);Intros.
Split;Intros.
Split;Intros.
Generalize (from_to2_Zlength `(bsp_p tt)-1` H);Intro.
Auto with zarith.
Apply from_to2_nth;Auto.
Absurd  `i <> root`;Auto.
Split;Intros;Auto.
Absurd `i = root`;Auto.
Defined.


Definition gather_list : (T:Set) (vec:(Vector T)) (root:Z) `0<=root<(bsp_p tt)` -> 
 { res: (Vector (list T)) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
  ((i=root) -> ( (Zlength T (at ? res i (Hyp_i)))=(bsp_p tt) /\ (j:Z) 
   (Hyp_j:`0<=j<(bsp_p tt)`)
     (Znth_spec T (at ? res i (Hyp_i)) `j+1` (at T vec j Hyp_j)))) /\ ( (`i<>root`) -> (at ? res i (Hyp_i))=(nil T))}.
Intros.
Refine Cases (procs_lists_gather root) of
  (exist p_proc h_proc) => Cases (gather T root vec) of
    (exist p_gather h_gather) => Cases (parfun ? ? (compose Z (option T) T (noSome T)) p_gather ) of
      (exist to_list0 h_to_list) => Cases (parfun2 (Z->T) (list Z) (list T) (map 1!Z 2!T) to_list0 p_proc) of
	 (exist res hres) => (exist ? ? res ?) end end end end.

Intros.
Rewrite (hres i Hyp_i).
Split;Intros.
Split;Intros.

Generalize (h_proc i Hyp_i);Intro.
Inversion H1.
Generalize (H2 H0);Intro.
Inversion H4.
Generalize (Zlength_map Z T (at (list Z) p_proc i Hyp_i) (at (Z->T) to_list0 i Hyp_i));Intro.
Rewrite H7.
Rewrite H5;Auto.

Rewrite (h_to_list i Hyp_i).
Generalize (h_gather i Hyp_i);Intro.
Inversion H1;Clear H1 H3.
Generalize (H2 H0);Intro;Clear H2.
Generalize (H1 j);Intro;Clear H1.
Inversion H2;Clear H2 H3.
Generalize (H1 Hyp_j);Intro;Clear H1.
Generalize (h_proc i Hyp_i);Intro.
Inversion H1;Clear H1 H4.
Generalize (H3 H0);Intro;Clear H3.
Inversion H1;Clear H1 H3.
Assert `0 <= j <= (bsp_p tt)-1`.
Apply bidon6;Auto.
Generalize (H4 j H1);Intro;Clear H1 H4.
Unfold compose.
Apply (map_Znth2 Z T  (at (list Z) p_proc i Hyp_i) 
 ([x:Z](noSome T (at Z->(option T) p_gather i Hyp_i x))) j (at T vec j Hyp_j) );Auto.
Rewrite H2.
Simpl;Auto.

Generalize (h_proc i Hyp_i);Intro.
Inversion H1;Clear H2 H1.
Generalize (H3 H0);Intro;Clear H3.
Rewrite H1;Simpl;Auto.
Defined.


(** ** Scan algorithms *)

Inductive k_first_R [T:Set] [R:T->T->T] [v:(Vector T)]
               :(k:Z) `0<=k<(bsp_p tt)` -> (res:T) Prop:=
   k_zero: (k_first_R T R v `0` zero_within_bound (at T v `0` zero_within_bound))
 | k_rec : (k:Z) (H_r:`0<=k<(bsp_p tt)`) (a:T) (H_sub:`k>0`)
                 (k_first_R T R v `k-1` (p_sub k H_r H_sub)  a) 
             -> (k_first_R T R v `k` H_r (R a (at T v `k` H_r))).

Axiom fold_left_k_first : (T:Set) (op:T->T->T)  (vec:(Vector T)) (hd:T) (hd=(at T vec `0` zero_within_bound)) ->
(taille:Z) (Hyp_tai:`0<=taille<(bsp_p tt)`) (l:(list T)) (((Zlength T l)=taille) /\ (pos:Z) (Hyp_pos:`0<pos<taille`)  (Znth_spec T l `pos` (at T vec pos (within_bound_k_first taille Hyp_tai pos Hyp_pos)))) ->
 (k_first_R T op vec taille Hyp_tai (fold_left op l hd)).

(** *** Scan direct *)

(** Create a parallel vector of list where the elements of the list are the name of the processes *)
Definition procs_lists : unit -> {res : (Vector (list Z)) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
(Zlength Z (at ? res i Hyp_i))=`i+1` /\ (pos:Z) `0<=pos<=i` -> (Znth_spec Z (at ? res i Hyp_i) `pos+1` pos)}.
Intros.
Exists (mkpar ? ([pid:Z](from_to2 pid))).
Intros.
Rewrite (mkpar_def);Auto.
Split;Intros.
Apply from_to2_Zlength;Auto.
Auto with zarith.
Apply from_to2_nth;Auto.
Auto with zarith.
Defined.

Definition scan_direct_tmp1: (T:Set) (vec:(Vector T)) {res: (Vector (list T)) | (i:Z) (Hyp_i:`0 <= i < (bsp_p tt)`)
(Zlength  T (at (list T) res i Hyp_i)) = `i+1` /\
   (pos:Z) (Hyp_pos:`0 <= pos <=i`) (Znth_spec T (at (list T) res i Hyp_i) `pos+1` (at T vec pos (scan_bound1 i Hyp_i pos Hyp_pos)))}.
Intros.
Refine let mkmsg = ([pid:Z] [v:T] [dst:Z] (if (Z_lt_dec dst pid) then [H1] (None T) 
                                            else [H2] (Some T v))) in
       let receivedmsgs = (put ? (apply ? ? (mkpar ? mkmsg) vec)) in
         Cases (parfun ? ? (compose Z (option T) T (noSome T)) receivedmsgs ) of
          (exist p h) => Cases (procs_lists tt) of 
              (exist p_proc h_proc) => Cases (parfun2 (Z->T) (list Z) (list T) (map 1!Z 2!T) p p_proc) of
	        (exist res hres) => (exist ? ? res ?) end end end.

Intros.
Rewrite (hres i Hyp_i).
Split;Intros.
Generalize (h_proc i Hyp_i);Intro.
Inversion H.
Generalize (Zlength_map Z T (at (list Z) p_proc i Hyp_i) (at (Z->T) p i Hyp_i) );Intro.
Rewrite H2;Auto.
Assert (Znth_spec Z (at (list Z) p_proc i Hyp_i) `pos+1` pos).
Generalize (h_proc i Hyp_i);Intro.
Inversion H.
Apply (H1 pos Hyp_pos).
Assert  (at T vec pos (scan_bound1 i Hyp_i pos Hyp_pos))=( (at Z->T p i Hyp_i) pos).
Rewrite (h i Hyp_i).
Unfold compose.
Unfold  receivedmsgs.
Rewrite put_def;Auto.
Case (within_bound pos);Intros.
Rewrite apply_def;Auto.
Rewrite mkpar_def;Auto.
Unfold mkmsg.
Case (Z_lt_dec i pos);Intros.
Generalize (scan_direct_lem1 i pos Hyp_pos z);Intro.
Elim H0.
Simpl.
Rewrite (at_H T vec pos  (scan_bound1 i Hyp_i pos Hyp_pos) a);Auto.
Generalize (scan_direct_lem2 i pos Hyp_i Hyp_pos  n);Intro.
Elim H0.
Apply (map_Znth2 Z T  (at (list Z) p_proc i Hyp_i)  (at Z->T p i Hyp_i) pos (at T vec pos (scan_bound1 i Hyp_i pos Hyp_pos)) H0 `pos+1` H).
Defined.


Lemma fold_left_tmp2: (T:Set) (l:(list T))(a,b:T) (Znth_spec T (cons a l) `1` b) -> a=b.
Intros.
Inversion H;Auto.
Generalize (Znth_lt_O T l n b H4);Intro.
Absurd `n+1 = 1`;Auto.
Omega.
Qed.

Lemma fold_left_tmp3: (T:Set) (l:(list T))(a,b:T) (pos:Z) `0<pos` -> (Znth_spec T (cons a l) `pos+1` b) -> (Znth_spec T l `pos` b).
Intros. 
Inversion H0.
Absurd `1 = pos+1`;Auto.
Omega.
Assert pos=n.
Auto with zarith.
Rewrite H6;Auto.
Qed.

Definition scan_direct : (T:Set) (op:T->T->T) (vec:(Vector T)) { res :(Vector T) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`)
                         (k_first_R T op vec i Hyp_i (at T res i Hyp_i))}.
Intros.
Refine Cases (scan_direct_tmp1 T vec) of
 (exist p h) => Cases (tl_hd_vect T p ?) of 
    (exist p' h') => Cases (parfun ? ? ( [hd_tl:(T*(list T))] (fold_left op (Snd hd_tl) (Fst hd_tl))) p') of
      (exist p'' h'') => (exist ? ? p'' ?) end end end;Intros.

Generalize (h i Hyp_i);Intro.
Inversion H.
Clear H.
Assert `0<i+1`.
Apply bidon2;Auto.
Apply (Zlength_no_nil T (at (list T) p i Hyp_i) `i+1` H H0).

Rewrite (h'' i Hyp_i).
Generalize (h' i Hyp_i);Intro.
Generalize (h i Hyp_i);Intro.
Inversion H0.

Assert  (Fst (at T*(list T) p' i Hyp_i))=(at T vec `0` zero_within_bound).
Assert (at (list T) p i Hyp_i)=(cons (Fst (at T*(list T) p' i Hyp_i)) (Snd (at T*(list T) p' i Hyp_i))).
Auto.
Assert `0<=0<=i`.
Apply bidon3;Auto.
Generalize (H2 `0` H4);Rewrite H3;Intro.
Generalize (fold_left_tmp2 T  (Snd (at T*(list T) p' i Hyp_i)) (Fst (at T*(list T) p' i Hyp_i)) (at T vec `0` (scan_bound1 i Hyp_i `0` H4)) H5);Intro.
Assert (at T vec `0` zero_within_bound)=(at T vec `0` (scan_bound1 i Hyp_i `0` H4)).
Apply (at_H T vec `0` zero_within_bound (scan_bound1 i Hyp_i `0` H4)).
Rewrite H7;Auto.

Assert `(Zlength T (snd T (list T) (at (prod T (list T)) p' i Hyp_i)))= i` /\
  ((pos:Z; Hyp_pos:`0 < pos < i`) (Znth_spec T (Snd (at T*(list T) p' i Hyp_i)) pos (at T vec pos (within_bound_k_first i Hyp_i pos Hyp_pos)))).

Split.
Assert (at (list T) p i Hyp_i)=(cons (Fst (at T*(list T) p' i Hyp_i)) (Snd (at T*(list T) p' i Hyp_i))).
Auto.
Generalize H1;Rewrite H4;Simpl;Intro.
Apply (bidon4);Auto.

Intros.
Assert `0<pos`.
Inversion Hyp_pos;Auto.
Assert (at (list T) p i Hyp_i)=(cons (Fst (at T*(list T) p' i Hyp_i)) (Snd (at T*(list T) p' i Hyp_i))).
Auto.
Generalize (H2 pos (within_bound_k_first2 i Hyp_i pos Hyp_pos)).
Rewrite H5;Intro.
Assert (at T vec pos (scan_bound1 i Hyp_i pos (within_bound_k_first2 i Hyp_i pos Hyp_pos)))= (at T vec pos (within_bound_k_first i Hyp_i pos Hyp_pos)).
Apply (at_H T vec pos).
Generalize H6.
Rewrite H7;Intro.
Apply (fold_left_tmp3 T (Snd (at T*(list T) p' i Hyp_i)) (Fst (at T*(list T) p' i Hyp_i)) (at T vec pos (within_bound_k_first i Hyp_i pos Hyp_pos)) pos H4 H8).

Apply (fold_left_k_first T op vec (Fst (at T*(list T) p' i Hyp_i)) H3 i Hyp_i (Snd (at T*(list T) p' i Hyp_i)) H4).
Defined.


(** *** Scan list *)

(** *** Divide-and-conquer Superposition Scan *)

(** *** Logarithmic scan *)

(** ** Fold_left *)


Definition fold_left_vec : (T:Set) (op:T->T->T) (vec:(Vector T))
 { res: (Vector T) | (i:Z) (Hyp_i:`0<=i<(bsp_p tt)`) (k_first_R T op vec (last tt) last_within_bound (at T res i Hyp_i))}.
Intros.      
Refine Cases (total_exchange T vec) of
    (exist p h) => Cases (tl_hd_vect T p ?) of
      (exist p' h') => Cases (parfun ? ? ( [hd_tl:(T*(list T))] (fold_left op (Snd hd_tl) (Fst hd_tl))) p') of
        (exist p'' h'') => (exist ? ? p'' ?) end end end;Intros.

Generalize (h i Hyp_i);Intro.
Inversion H.
Apply (Zlength_no_nil T (at (list T) p i Hyp_i) (bsp_p tt) good_bsp_p H0).
Rewrite (h'' i Hyp_i).
Generalize (h' i Hyp_i);Intro.
Generalize (h i Hyp_i);Intro.
Inversion H0.

Assert (Zlength T (Snd (at T*(list T) p' i Hyp_i)))=`(bsp_p tt)-1`.
Assert (at (list T) p i Hyp_i)=(cons (Fst (at T*(list T) p' i Hyp_i)) (Snd (at T*(list T) p' i Hyp_i))).
Auto.
Generalize H1;Rewrite H3;Simpl;Intro.
Apply (bidon1);Auto.

Assert  (Fst (at T*(list T) p' i Hyp_i))=(at T vec `0` zero_within_bound).
Assert (at (list T) p i Hyp_i)=(cons (Fst (at T*(list T) p' i Hyp_i)) (Snd (at T*(list T) p' i Hyp_i))).
Auto.
Generalize H2;Rewrite H4;Intro.
Generalize (H5 `0` zero_within_bound);Intro.
Apply (fold_left_tmp2 T  (Snd (at T*(list T) p' i Hyp_i)) (Fst (at T*(list T) p' i Hyp_i)) (at T vec `0` zero_within_bound) H6).

Assert ((pos:Z; Hyp_pos:`0 < pos < (last tt)`) (Znth_spec T (Snd (at T*(list T) p' i Hyp_i)) pos
             (at T vec pos (within_bound_k_first (last tt) last_within_bound pos Hyp_pos)))).
Intros.
Assert `0<pos`.
Inversion Hyp_pos;Auto.
Assert (at (list T) p i Hyp_i)=(cons (Fst (at T*(list T) p' i Hyp_i)) (Snd (at T*(list T) p' i Hyp_i))).
Auto.
Generalize (H2 pos (within_bound_k_first (last tt) last_within_bound pos Hyp_pos)).
Rewrite H6;Intro.
Apply (fold_left_tmp3 T (Snd (at T*(list T) p' i Hyp_i)) (Fst (at T*(list T) p' i Hyp_i)) (at T vec pos (within_bound_k_first (last tt) last_within_bound pos Hyp_pos)) pos H5);Auto.

Assert `(Zlength T (snd T (list T) (at (prod T (list T)) p' i Hyp_i))) = (last tt)`
       /\ ((pos:Z; Hyp_pos:`0 < pos < (last tt)`) (Znth_spec T (Snd (at T*(list T) p' i Hyp_i)) pos
             (at T vec pos (within_bound_k_first (last tt) last_within_bound pos Hyp_pos)))).
Split; [ Unfold last;Auto | Auto ].
Apply (fold_left_k_first T op vec (Fst (at T*(list T) p' i Hyp_i)) H4 (last tt) last_within_bound (Snd (at T*(list T) p' i Hyp_i)) H6).
Defined.
