(* ########################################################################## *)
(* ############################### bsmllemma  ############################### *)
(* ########################################################################## *)

(** Some technical lemmas for the proofs. *)

Require ZArith.
Require Coq.ZArith.Zdiv.
Require Omega.
Require PolyList.
Require Bsml.

Require Export ZArith.
Require Export Coq.ZArith.Zdiv.
Require Export Omega.
Require Export PolyList.
Require Export Bsml.

(** *** Other lemmas *)

(** Irreflexivity of the parallel vectors *)
Lemma at_egal: (T:Set)(a:Z) (H_a:`0<=a<(bsp_p tt)`) (b:Z) (H_b:`0<=b<(bsp_p tt)`) 
               (data:(Vector T)) (a=b) -> ((at T data a H_a)=(at T data b H_b)).
Intros.
NewInduction H.
Apply (at_H T data a H_a H_b).
Qed.


(** Reflexivity of the equality *)
Lemma egal_reflexive: (T:Set)(a,b:T) (a=b)->(b=a).
Intros.
Rewrite H;Auto.
Qed.


(** *** Equality with "Option" *)
Lemma egal_some: (T:Set)(a,b:T) ((Some ? a)=(Some ? b)) -> (a=b).
Intros.
Inversion H;Auto.
Qed.

Lemma some_egal:(T:Set)(a,b:T) (a=b) -> ((Some ? a)=(Some ? b)).
Intros.
Induction H;Auto.
Qed.

Lemma no_None_so_Some: (T:Set) (p:(option T)) ~(p=(None T)) -> (Ex [x:T] p=(Some T x)).
Intros T p.
Case p;Intros.
Exists t;Auto.
Absurd ~(None T)=(None T); Auto.
Qed.


(** *** Equality with "List" *)

Lemma no_nil_so_hd_tl: (T:Set) (l:(list T)) ~(l=(nil T)) -> (Ex [x0:T] (Ex [l0:(list T)] l=(cons x0 l0))).
Intros T l.
Case l;Intros.
Absurd  ~(nil T)=(nil T);Auto.
Exists t.
Exists l0;Auto.
Qed.


(** *** Bound and arith *)

Lemma p_sub : (k:Z) `0<=k<(bsp_p tt)` -> `k>0` -> `0<=k-1<(bsp_p tt)`.
Intros.
Auto with zarith.
Qed.

Lemma zero_within_bound: `0<=0<(bsp_p tt)`.
Auto with zarith.
Qed.

Lemma within_bound_lt: (n:Z) (i:Z) `0 < i < n` -> `0 <= i < n`.
Intros.
Auto with zarith.
Qed.

Remark bidon1: (a,b:Z) `a+1=b` -> `a=b-1`.
Intros.
Auto with zarith.
Qed.


Remark bidon2: (a:Z) `0<=a<(bsp_p tt)` -> `0<a+1`.
Intros.
Auto with zarith.
Qed.

Remark bidon3: (i:Z) `0<=i<(bsp_p tt)` -> `0<=0<=i`.
Intros.
Auto with zarith.
Qed.

Remark bidon4: (a,b:Z) `a+1=b+1` -> a=b.
Intros.
Auto with zarith.
Qed.

Remark bidon5: (n:Z) `1<=n` ->  `n=0` -> `n<>0`.
Auto with zarith.
Qed.

Remark bidon6: (j:Z) `0 <= j < (bsp_p tt)` -> `0 <= j <= (bsp_p tt)-1`.
Intros.
Auto with zarith.
Qed.

Remark within_bound_k_first: (n:Z) `0<=n<(bsp_p tt)` -> (i:Z) `0<i<n` -> `0<=i<(bsp_p tt)`.
Intros.
Auto with zarith.
Qed.

Remark k_first_tmp1 : (k:Z) `k=0` -> `0<=k<(bsp_p tt)`.
Intros.
Rewrite H;Auto with zarith.
Qed.

Lemma within_bound_k_first2 : (n:Z) `0 <= n < (bsp_p tt)`->(i:Z)`0 < i < n`->`0 <= i <= n`.
Intros.
Auto with zarith.
Qed.

Lemma scan_bound1: (i:Z) `0 <= i < (bsp_p tt)` -> (pos:Z) `0 <= pos <=i` -> `0 <= pos < (bsp_p tt)`.
Intros.
Auto with zarith.
Qed.

Lemma scan_direct_lem2: (i:Z)(pos:Z) `0 <= i < (bsp_p tt)` -> `0 <= pos <= i` ->  ~`0 <= pos < (bsp_p tt)` -> False.
Intros.
Omega.
Qed.

Lemma scan_direct_lem1: (i:Z)(pos:Z) `0 <= pos <= i` ->  `i < pos` -> False.
Intros.
Omega.
Qed.



(** arith with modulo *)
Lemma shift_tmp1:(n:Z) `0<= n%(bsp_p tt) < (bsp_p tt)`.
Intro.
Apply (Z_mod_lt n (bsp_p tt)).
Generalize good_bsp_p;Auto with zarith.
Qed.

Lemma shift_dec: (dec:Z) (i:Z) `0<=i<(bsp_p tt)` -> `0<=((bsp_p tt)+i-(dec%(bsp_p tt)))% (bsp_p tt)< (bsp_p tt)`.
Intros.
Apply (shift_tmp1 `((bsp_p tt)+i-dec%(bsp_p tt))`).
Qed.


(** have to proof in the futur *)
Axiom shift_left1:`((bsp_p tt)+0-1%(bsp_p tt))%(bsp_p tt) = (bsp_p tt)-1`.
Axiom shift_left2:(i:Z)`0 < i < (bsp_p tt)` -> `((bsp_p tt)+i-1%(bsp_p tt))%(bsp_p tt) = i-1`.
Axiom shift_right1: `((bsp_p tt)+((bsp_p tt)-1)-(-1)%(bsp_p tt))%(bsp_p tt) = 0`.
