Require PolyList.
Require Bsml.

Require Export PolyList.
Require Export Bsml.

(** * Definition of a local_sort. *)
(** From the works of Catherine Parent *)
Section Local_sort.

Variable A:Set.

(* A decidable relation on A *)
Variable   inf : A -> A -> Prop.
Definition sup  :=[x,y:A]~(inf x y).

Hypothesis inf_sup : (x,y:A){(inf x y)}+{(sup x y)}.
Hypothesis inf_sup_sym : (x,y:A)(sup x y)->(inf y x).
Hypothesis inf_tran : (x,y,z:A)(inf x y)->(inf y z)->(inf x z).

Hints Immediate inf_sup_sym : progs.

Lemma inf_sup_tran : (x,y,z:A)(sup y x)->(inf y z)->(inf x z).
Intros x y z H H0 ; Apply inf_tran with y ; Auto.
Qed.


(* Equivalence of two lists  *)
Inductive permut : (list A)->(list A)->Prop :=
    permut_refl  : (l:(list A))(permut l l)
   |permut_tran  : (l,m,n:(list A))(permut l m)->(permut m n)->(permut l n)
   |permut_mil   : (a:A)(l,m:(list A))
                   (permut (cons a (app l m)) (app l (cons a m)))
   |permut_app   : (l,l',m,m':(list A))(permut l l')->(permut m m')->
                                   (permut (app l m) (app l' m')).

Hints Resolve permut_refl permut_mil permut_app : progs.

(* if l eq m then a:l eq a:m  *)
Lemma permut_cons : (a:A)(l,m:(list A))(permut l m)->(permut (cons a l) (cons a m)).
Intros a l m H.
Change (permut (app (cons a (nil A)) l) (app (cons a (nil A)) m)).
Auto with progs.
Qed.


(* if l eq m then (incl l m) and (incl m l) *)
Lemma permut_in : (m,n:(list A))(permut m n)->(incl m n).
Intros m n H ; Elim H ; Auto with v62 progs.
Intros;Apply(incl_tran H1 H3).
Qed.
Hints Immediate permut_in : progs.

Lemma permut_in_sym : (m,n:(list A))(permut m n)->(incl n m).
Intros m n H ; Elim H ; Auto with v62 progs.
Intros;Apply(incl_tran H3 H1).
Qed.
Hints Immediate permut_in_sym : progs.

(* Each element of a list satisfies a given predicate R *)
Section R_on_list.

Variable R : A->Prop.

Definition Rlist := [l:(list A)](a:A)(In a l)->(R a).

Lemma Rnil : (Rlist (nil A)).
Red; Intros a H.
Elim H.
Qed.
Hints Resolve Rnil : progs.


Lemma Rcons : (x:A)(l:(list A))(R x)->(Rlist l)->(Rlist (cons x l)).
Unfold Rlist ; Simpl ; Intros x l H H0 a H1.
Elim H1 ; Auto with v62 progs ; Intro H2.
Elim H2 ; Auto with v62 progs.
Qed.
Hints Resolve Rcons : progs.

Lemma Rlist_app : (m,n:(list A))(Rlist m)->(Rlist n)->(Rlist (app m n)).
Unfold Rlist ; Intros m n H H0 a H1.
Elim (in_app_or H1) ; Auto with v62 progs.
Qed.
Hints Resolve Rlist_app : progs.


Lemma Rincl : (m,n: (list A))(incl m n)->(Rlist n)->(Rlist m).
Unfold Rlist incl ; Auto with v62 progs.
Qed.
Hints Resolve Rincl : progs.

End R_on_list.
Hints Resolve Rnil Rcons Rlist_app : progs.

(* particularisation to some predicates    *)
(* (inf_list x l) == (inf x a) for a in l  *)
(* (sup_list x l) == (sup x a) for a in l  *)
(* (infl l m) == (inf_list a m) for a in l *)

Definition inf_list := [x:A](Rlist (inf x)).
Definition sup_list := [x:A](Rlist (sup x)).
Definition infl := [l,m:(list A)](Rlist [a:A](inf_list a m) l).

Hints Unfold   inf_list sup_list infl : progs.

(* properties of these relations *)

(* l <= nil *)
Lemma infl_nil : (l:(list A))(infl l (nil A)).
Unfold infl Rlist ; Auto with v62 progs.
Qed.
Hints Resolve infl_nil : progs.

(*   (a > m) => (m <= l) => (m <= a:l)  *)
Lemma infl_cons : (a:A)(l,m:(list A))(sup_list a m)->(infl m l)->(infl m (cons a l)).
Unfold infl inf_list sup_list Rlist ; Simpl.
Intros a  l m H H0 a0 H1 a1 H2.
Elim H2 ; Auto with v62 progs ; Intro H3.
Elim H3 ; Apply inf_sup_sym ; Auto with v62 progs.
Qed.
Hints Resolve infl_cons : progs.


(*   (a > m) => (a <= l) => (m <= l) *)
Lemma sup_inf_infl : (a:A)(l,m:(list A))(sup_list a l)->(inf_list a m)->(infl l m).
Unfold infl inf_list sup_list Rlist ; Intros a l m H H0 a0 H1 a1 H2.
Apply inf_sup_tran with a ; Auto with v62 progs.
Qed.



(* Definition for a list to be sorted *)
Inductive sort: (list A)->Prop :=
     sort_nil : (sort (nil A))
   | sort_cons : (a:A)(l:(list A))(inf_list a l)->(sort l)->(sort (cons a l)).

Hints Resolve sort_nil sort_cons : progs.


End Local_sort.

Section Sample_Sorted.

Variable T:Set.
Variable inf : T -> T -> Prop.   

Lemma p_next : (pos:Z)  `0 <= pos < (bsp_p tt)` -> `pos<>(bsp_p tt)-1` -> `0 <= pos+1 < (bsp_p tt)`.
Auto with zarith.
Qed.

Inductive PSRS_sorted [v:(Vector (list T))] : Prop :=
 PSRS_sor: (i:Z) (H:`0<=i<(bsp_p tt)`) ((sort T inf (at ? v i H)) /\ 
           ((H':`i<>(bsp_p tt)-1`) (infl T inf (at ? v i H) (at ? v `i+1` (p_next i H H'))))) -> (PSRS_sorted v).
  

Inductive PSRS_exchange [v1,v2:(Vector (list T))] : Prop :=
    PSRS_exch : (i:Z) (H_i:`0<=i<(bsp_p tt)`) (l:(list T)) 
 (incl l (at ? v1 i H_i)) -> (Ex [j:Z] (H_j:`0<=j<(bsp_p tt)`) 
    (incl l (at ? v1 j H_j))) -> (PSRS_exchange v1 v2).
  

Inductive PSRS_permut : (Vector (list T))->(Vector (list T)) -> Prop :=
  PSRS_exch_is_permut : 
    (v,v':(Vector (list T))) (PSRS_exchange v v') -> (PSRS_permut v v')
   | PSRS_permut_refl :  (v:(Vector (list T)))(PSRS_permut v v)
   | PSRS_permut_sym : (v,v':(Vector (list T)))(PSRS_permut v v') 
                        -> (PSRS_permut v' v)
   | PSRS_permut_trans : (v,v',v'':(Vector (list T))) (PSRS_permut v v')
        -> (PSRS_permut v' v'') -> (PSRS_permut v v'').

End Sample_Sorted.
